from spidermon.contrib.stats.analyzer import StatsAnalyzer
from spidermon.contrib.stats.counters import DictPercentCounter, PercentCounter
from spidermon.exceptions import NotConfigured

from .job import JobMonitorMixin
from .stats import StatsMonitorMixin

DOWNLOADER_RESPONSE_COUNT = "downloader/response_count"
DOWNLOADER_RESPONSE_STATUS = "downloader/response_status_count/"
DOWNLOADER_STATUS_CODES_INFORMATIONAL = [r"1\d{2}$"]
DOWNLOADER_STATUS_CODES_SUCCESSFUL = [r"2\d{2}$"]
DOWNLOADER_STATUS_CODES_REDIRECTIONS = [r"3\d{2}$"]
DOWNLOADER_STATUS_CODES_BAD_REQUESTS = [r"4\d{2}$"]
DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS = [r"5\d{2}$"]
DOWNLOADER_STATUS_CODES_OTHERS = ["[^1-5].*$"]
DOWNLOADER_STATUS_CODES_ERRORS = (
    DOWNLOADER_STATUS_CODES_BAD_REQUESTS
    + DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS
)


class ResponsesInfo:
    def __init__(self, stats):
        self._stats_analyzer = StatsAnalyzer(stats=stats)
        self.count = self._stats_analyzer.search(DOWNLOADER_RESPONSE_COUNT + "$").get(
            DOWNLOADER_RESPONSE_COUNT, 0
        )

        # all status codes
        self.all = DictPercentCounter(total=self.count)
        self._add_status_codes(pattern=None, target=self.all)

        # 1xx. informational
        self.informational = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_INFORMATIONAL, target=self.informational
        )

        # 2xx. successful
        self.successful = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_SUCCESSFUL, target=self.successful
        )

        # 3xx. redirections
        self.redirections = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_REDIRECTIONS, target=self.redirections
        )

        # 4xx. bad requests
        self.bad_requests = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_BAD_REQUESTS, target=self.bad_requests
        )

        # 5xx. internal server errors
        self.internal_server_errors = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS,
            target=self.internal_server_errors,
        )

        # >= 6xx. others
        self.others = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_OTHERS, target=self.others
        )

        # errors (4xx + 5xx)
        self.errors = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_ERRORS, target=self.errors
        )

    def _add_status_codes(self, pattern, target):
        for code, counter in self._get_response_codes(pattern).items():
            target.add_value(code, counter.count)

    def _get_response_codes(self, codes=None):
        codes = codes or ["[^/]+"]
        return_codes = {}
        for code in codes:
            return_codes.update(self._get_response_code(code))
        return return_codes

    def _get_response_code(self, code):
        return {
            code: PercentCounter(count, self.count)
            for count, code in self._stats_analyzer.search(
                pattern=DOWNLOADER_RESPONSE_STATUS + ("(%s)$" % code),
                include_matches=True,
            ).values()
        }


class SpiderMonitorMixin(StatsMonitorMixin, JobMonitorMixin):
    @property
    def crawler(self):
        if not self.data.crawler:
            raise NotConfigured("Crawler not available!")
        return self.data.crawler

    @property
    def spider(self):
        if not self.data.spider:
            raise NotConfigured("Spider not available!")
        return self.data.spider

    @property
    def responses(self):
        if not hasattr(self, "_responses"):
            self._responses = ResponsesInfo(self.stats)
        return self._responses



import os
import inspect

from jinja2 import Template
from spidermon.core.actions import ActionOptionsMetaclass, Action
from spidermon.templates import template_loader


class ActionWithTemplatesMetaclass(ActionOptionsMetaclass):
    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)
        mcs.add_class_templates(cls)
        return cls

    @classmethod
    def add_class_templates(mcs, cls):
        class_file = inspect.getfile(cls)
        class_path = os.path.dirname(class_file)
        template_loader.discover_folder(class_path)
        for path in cls.template_paths:
            template_path = os.path.join(class_path, path)
            template_loader.add_path(template_path)


class ActionWithTemplates(Action, metaclass=ActionWithTemplatesMetaclass):
    template_paths = []
    context = None

    def __init__(self, context=None):
        super().__init__()
        self.context = context or self.context or {}

    def get_template(self, name):
        return template_loader.get_template(name)

    def render_text_template(self, template):
        """Render a Jinja2 template given in *template* as a string.

        For example::

            action.render_text_template('{{ monitors_failed }} monitors failed!')
        """
        template = Template(template)
        return template.render(self.get_template_context())

    def render_template(self, template):
        """Render the Jinja2 template file named *template*.

        It uses :data:`spidermon.templates.template_loader` to resolve
        *template* to an actual template file.

        For example::

            action.render_template('mytemplate.html')
        """
        template = self.get_template(template)
        return template.render(self.get_template_context())

    def get_template_context(self):
        context = {
            "result": self.result,
            "data": self.data,
            "monitors_passed": self.monitors_passed,
            "monitors_failed": self.monitors_failed,
        }
        context.update(self.context)
        return context



"""
Module to hold a reference to singleton Hubstorage client and Job instance
"""
import os
from codecs import decode


try:
    try:
        from scrapinghub import HubstorageClient
    except ImportError:
        from hubstorage.client import HubstorageClient
except ImportError:
    HubstorageClient = None


class _Hubstorage:
    def __init__(self):
        self.available = "SHUB_JOBKEY" in os.environ and HubstorageClient is not None
        self._client = None
        self._project = None
        self._job = None
        if self.available:
            self.job_key = os.environ["SHUB_JOBKEY"]
            self._project_id, self._spider_id, self._job_id = map(
                int, self.job_key.split("/")
            )
        else:
            self._project_id = None
            self._spider_id = None
            self._job_id = None

    @property
    def auth(self):
        return decode(os.environ["SHUB_JOBAUTH"], "hex_codec").decode("utf-8")

    @property
    def endpoint(self):
        return os.environ.get("SHUB_STORAGE")

    @property
    def project_id(self):
        return self._project_id

    @property
    def spider_id(self):
        return self._spider_id

    @property
    def job_id(self):
        return self._job_id

    @property
    def client(self):
        if self._client is None:
            self._client = HubstorageClient(endpoint=self.endpoint, auth=self.auth)
        return self._client

    @property
    def project(self):
        if self._project is None:
            self._project = self.client.get_project(str(self.project_id))
        return self._project

    @property
    def job(self):
        if self._job is None:
            self._job = self.project.get_job((self.spider_id, self.job_id))
        return self._job

    def close(self):
        if self._client is not None:
            self._client.close()


hs = _Hubstorage()



import collections


class PercentCounterBase:
    def __init__(self, total=0):
        self._total = total

    @property
    def count(self):
        raise NotImplementedError

    @property
    def percent(self):
        if self._total <= 0 or self.count <= 0:
            return 0
        else:
            return float(self.count) / float(self._total)

    def __str__(self):
        return "(count=%d, percent=%.2f)" % (self.count, self.percent)

    def __repr__(self):
        return self.__str__()


class PercentCounter(PercentCounterBase):
    def __init__(self, count=0, total=0):
        super().__init__(total)
        self._count = count

    @property
    def count(self):
        return self._count

    def inc_value(self, value):
        self._count += value


class DictPercentCounter(PercentCounterBase, collections.abc.MutableMapping):
    __items_class__ = PercentCounter

    def __init__(self, total):
        super().__init__(total)
        self._dict = dict()

    @property
    def count(self):
        return sum([e.count for e in self._dict.values()])

    def add_value(self, key, value):
        if key not in self._dict:
            self._create_item(key)
        self[key].inc_value(value)

    def _create_item(self, key):
        self._dict[key] = self.__items_class__(total=self._total)

    def __getitem__(self, key):
        if key not in self._dict:
            return self.__items_class__(total=self._total)
        else:
            return self._dict[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __keytransform__(self, key):
        return key

    def _immutable(self, *args, **kws):
        raise TypeError

    def __str__(self):
        return "(count=%d, percent=%.2f, %s)" % (
            self.count,
            self.percent,
            str(self._dict),
        )

    __setitem__ = _immutable
    __delitem__ = _immutable


class AttributeDictPercentCounter(PercentCounterBase):
    __attribute_dict_name__ = "dict"

    def __init__(self, total):
        super().__init__(total)
        setattr(self, self.__attribute_dict_name__, DictPercentCounter(total))

    @property
    def attribute_dict(self):
        return getattr(self, self.__attribute_dict_name__)

    @property
    def count(self):
        return sum([e.count for e in self.attribute_dict.values()])

    def add_value(self, key, value):
        self.attribute_dict.add_value(key, value)

    def __str__(self):
        return "(count=%d, percent=%.2f, %s=%s)" % (
            self.count,
            self.percent,
            self.__attribute_dict_name__,
            str(self.attribute_dict),
        )

    def __repr__(self):
        return self.__str__()



from spidermon import Monitor, MonitorSuite, monitors
from spidermon.exceptions import NotConfigured
from spidermon.utils.settings import getdictorlist

from ..monitors.mixins.spider import SpiderMonitorMixin

SPIDERMON_EXPECTED_FINISH_REASONS = "SPIDERMON_EXPECTED_FINISH_REASONS"
SPIDERMON_UNWANTED_HTTP_CODES = "SPIDERMON_UNWANTED_HTTP_CODES"
SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT = "SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT"
SPIDERMON_MAX_ITEM_VALIDATION_ERRORS = "SPIDERMON_MAX_ITEM_VALIDATION_ERRORS"
SPIDERMON_MAX_RETRIES = "SPIDERMON_MAX_RETRIES"
SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS = "SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS"
SPIDERMON_MIN_SUCCESSFUL_REQUESTS = "SPIDERMON_MIN_SUCCESSFUL_REQUESTS"
SPIDERMON_MAX_REQUESTS_ALLOWED = "SPIDERMON_MAX_REQUESTS_ALLOWED"


class BaseScrapyMonitor(Monitor, SpiderMonitorMixin):
    longMessage = False

    @property
    def monitor_description(self):
        if self.__class__.__doc__:
            return self.__class__.__doc__.split("\n")[0]
        return super().monitor_description


class BaseStatMonitor(BaseScrapyMonitor):
    """Base Monitor class for stat-related monitors.

    Create a monitor class inheriting from this class to have a custom
    monitor that validates numerical stats from your job execution
    against a configurable threshold.

    As an example, we will create a new monitor that will check if the
    value obtained in a job stat 'numerical_job_statistic' is greater than
    or equal to the value configured in ``CUSTOM_STAT_THRESHOLD`` project
    setting:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "numerical_job_statistic"
            threshold_setting = "CUSTOM_STAT_THRESHOLD"
            assert_type = ">="

    For the ``assert_type`` property you can select one of the following:

    ==  =====================
    >   Greater than
    >=  Greater than or equal
    <   Less than
    <=  Less than or equal
    ==  Equal
    !=  Not equal
    ==  =====================

    Sometimes, we don't want a fixed threshold, but a dynamic based on more than
    one stat or getting data external from the job execution (e.g., you want the
    threshold to be related to another stat, or you want to get the value
    of a stat from a previous job).

    As an example, the following monitor will use as threshold the a variable number
    of errors allowed based on the number of items scraped. So this monitor will pass
    only if the number of errors is less than 1% of the number of items scraped:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "log_count/ERROR"
            assert_type = "<"

            def get_threshold(self):
                item_scraped_count = self.stats.get("item_scraped_count")
                return item_scraped_count * 0.01

    By default, if the stat can't be found in job statistics, the monitor will fail.
    If you want the monitor to be skipped in that case, you should set ``fail_if_stat_missing``
    attribute as ``False``.


    The following monitor will not fail if the job doesn't have a ``numerical_job_statistic``
    value in its statistics:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "numerical_job_statistic"
            threshold_setting = "CUSTOM_STAT_THRESHOLD"
            assert_type = ">="
            fail_if_stat_missing = False
    """

    fail_if_stat_missing = True

    def run(self, result):
        has_threshold_config = any(
            [hasattr(self, "threshold_setting"), hasattr(self, "get_threshold")]
        )
        if not has_threshold_config:
            raise NotConfigured(
                f"{self.__class__.__name__} should include a a `threshold_setting` attribute "
                "to be configured in your project settings with the desired threshold "
                "or a `get_threshold` method that returns the desired threshold."
            )

        if (
            hasattr(self, "threshold_setting")
            and self.threshold_setting not in self.crawler.settings.attributes
        ):
            raise NotConfigured(
                f"Configure {self.threshold_setting} to your project"
                f"settings to use {self.monitor_name}."
            )

        return super().run(result)

    def _get_threshold_value(self):
        if hasattr(self, "get_threshold"):
            return self.get_threshold()
        return self.crawler.settings.get(self.threshold_setting)

    def test_stat_monitor(self):
        assertions = {
            ">": self.assertGreater,
            ">=": self.assertGreaterEqual,
            "<": self.assertLess,
            "<=": self.assertLessEqual,
            "==": self.assertEqual,
            "!=": self.assertNotEqual,
        }
        threshold = self._get_threshold_value()

        if self.stat_name not in self.stats:
            message = f"Unable to find '{self.stat_name}' in job stats."
            if self.fail_if_stat_missing:
                self.fail(message)
            else:
                self.skipTest(message)

        value = self.stats.get(self.stat_name)

        assertion_method = assertions.get(self.assert_type)
        assertion_method(
            value,
            threshold,
            msg=f"Expecting '{self.stat_name}' to be '{self.assert_type}' "
            f"to '{threshold}'. Current value: '{value}'",
        )


@monitors.name("Extracted Items Monitor")
class ItemCountMonitor(BaseStatMonitor):
    """Check if spider extracted the minimum number of items.

    You can configure it using ``SPIDERMON_MIN_ITEMS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.
    """

    stat_name = "item_scraped_count"
    threshold_setting = "SPIDERMON_MIN_ITEMS"
    assert_type = ">="


@monitors.name("Critical Count Monitor")
class CriticalCountMonitor(BaseStatMonitor):
    """Check for critical errors in the spider log.

    You can configure it using ``SPIDERMON_MAX_CRITICALS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any critical error, the monitor will be skipped."""

    stat_name = "log_count/CRITICAL"
    threshold_setting = "SPIDERMON_MAX_CRITICALS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Error Count Monitor")
class ErrorCountMonitor(BaseStatMonitor):
    """Check for errors in the spider log.

    You can configure it using ``SPIDERMON_MAX_ERRORS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any error, the monitor will be skipped."""

    stat_name = "log_count/ERROR"
    threshold_setting = "SPIDERMON_MAX_ERRORS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Warning Count Monitor")
class WarningCountMonitor(BaseStatMonitor):
    """Check for warnings in the spider log.

    You can configure it using ``SPIDERMON_MAX_WARNINGS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any warning, the monitor will be skipped."""

    stat_name = "log_count/WARNING"
    threshold_setting = "SPIDERMON_MAX_WARNINGS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Finish Reason Monitor")
class FinishReasonMonitor(BaseScrapyMonitor):
    """Check if a job has a expected finish reason.

    You can configure the expected reason with the
    ``SPIDERMON_EXPECTED_FINISH_REASONS``, it should be an ``iterable`` of
    valid finish reasons.

    The default value of this settings is: ``['finished', ]``."""

    @monitors.name("Should have the expected finished reason(s)")
    def test_should_finish_with_expected_reason(self):
        expected_reasons = self.crawler.settings.getlist(
            SPIDERMON_EXPECTED_FINISH_REASONS, ("finished",)
        )
        finished_reason = self.stats.get("finish_reason")
        msg = 'Finished with "{}" the expected reasons are {}'.format(
            finished_reason, expected_reasons
        )
        self.assertTrue(finished_reason in expected_reasons, msg=msg)


@monitors.name("Unwanted HTTP codes monitor")
class UnwantedHTTPCodesMonitor(BaseScrapyMonitor):
    """Check for maximum number of unwanted HTTP codes.
    You can configure it using ``SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT`` setting
    or ``SPIDERMON_UNWANTED_HTTP_CODES`` setting

    This monitor fails if during the spider execution, we receive
    more than the number of ``SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT``
    setting for at least one of the HTTP Status Codes in the list defined in
    ``SPIDERMON_UNWANTED_HTTP_CODES`` setting.

    Default values are:

    .. highlight:: python
    .. code-block:: python

        SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT = 10
        SPIDERMON_UNWANTED_HTTP_CODES = [400, 407, 429, 500, 502, 503, 504, 523, 540, 541]

    ``SPIDERMON_UNWANTED_HTTP_CODES`` can also be a dictionary with the HTTP Status Code
    as key and the maximum number of accepted responses with that code.

    With the following setting, the monitor will fail if more than 100 responses are
    404 errors or at least one 500 error:

    .. highlight:: python
    .. code-block:: python

        SPIDERMON_UNWANTED_HTTP_CODES = {
            400: 100,
            500: 0,
        }

    """

    DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT = 10
    DEFAULT_UNWANTED_HTTP_CODES = [400, 407, 429, 500, 502, 503, 504, 523, 540, 541]

    @monitors.name("Should not hit the limit of unwanted http status")
    def test_check_unwanted_http_codes(self):
        unwanted_http_codes = getdictorlist(
            self.crawler,
            SPIDERMON_UNWANTED_HTTP_CODES,
            self.DEFAULT_UNWANTED_HTTP_CODES,
        )

        errors_max_count = self.crawler.settings.getint(
            SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT,
            self.DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT,
        )

        if not isinstance(unwanted_http_codes, dict):
            unwanted_http_codes = {
                code: errors_max_count for code in unwanted_http_codes
            }

        for code, max_errors in unwanted_http_codes.items():
            code = int(code)
            count = self.stats.get(f"downloader/response_status_count/{code}", 0)
            msg = (
                "Found {} Responses with status code={} - "
                "This exceed the limit of {}".format(count, code, max_errors)
            )
            self.assertTrue(count <= max_errors, msg=msg)


@monitors.name("Downloader Exceptions monitor")
class DownloaderExceptionMonitor(BaseScrapyMonitor):
    """Check the amount of downloader exceptions (timeouts, rejected
    connections, etc.).

    You can configure it using the ``SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS`` setting.
    The default is ``-1`` which disables the monitor.
    """

    @monitors.name("Should not hit the limit of downloader exceptions")
    def test_maximum_downloader_exceptions(self):
        exception_count = self.stats.get("downloader/exception_count", 0)
        threshold = self.crawler.settings.getint(
            SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS, -1
        )
        if threshold < 0:
            return
        msg = "Too many downloader exceptions ({})".format(exception_count)
        self.assertLessEqual(exception_count, threshold, msg=msg)


@monitors.name("Retry Count monitor")
class RetryCountMonitor(BaseScrapyMonitor):
    """Check if any requests have reached the maximum amount of retries
    and the crawler had to drop those requests.

    You can configure it using the ``SPIDERMON_MAX_RETRIES`` setting.
    The default is ``-1`` which disables the monitor.
    """

    @monitors.name(
        "Should not hit the limit of requests that reached the maximum retry amount"
    )
    def test_maximum_retries(self):
        max_reached = self.stats.get("retry/max_reached", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MAX_RETRIES, -1)
        if threshold < 0:
            return
        msg = "Too many requests ({}) reached the maximum retry amount".format(
            max_reached
        )
        self.assertLessEqual(max_reached, threshold, msg=msg)


@monitors.name("Successful Requests monitor")
class SuccessfulRequestsMonitor(BaseScrapyMonitor):
    """Check the amount of successful requests.

    You can configure it using the ``SPIDERMON_MIN_SUCCESSFUL_REQUESTS`` setting.
    """

    @monitors.name("Should have at least the minimum number of successful requests")
    def test_minimum_successful_requests(self):
        requests = self.stats.get("downloader/response_status_count/200", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MIN_SUCCESSFUL_REQUESTS, 0)
        msg = "Too few ({}) successful requests".format(requests)
        self.assertGreaterEqual(requests, threshold, msg=msg)


@monitors.name("Total Requests monitor")
class TotalRequestsMonitor(BaseScrapyMonitor):
    """Check the total amount of requests.

    You can configure it using the ``SPIDERMON_MAX_REQUESTS_ALLOWED`` setting.
    The default is ``-1`` which disables the monitor.
    """

    @monitors.name("Should not hit the total limit of requests")
    def test_request_count_exceeded_limit(self):
        requests = self.stats.get("downloader/request_count", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MAX_REQUESTS_ALLOWED, -1)
        if threshold < 0:
            return
        msg = "Too many ({}) requests".format(requests)
        self.assertLessEqual(requests, threshold, msg=msg)


@monitors.name("Item Validation Monitor")
class ItemValidationMonitor(BaseScrapyMonitor):
    """Check for item validation errors if item validation pipelines are enabled.

    You can configure the maximum number of item validation errors using
    ``SPIDERMON_MAX_ITEM_VALIDATION_ERRORS``. The default is ``0``."""

    @monitors.name("Should not have more item validation errors than configured.")
    def test_verify_item_validation_error(self):
        errors_threshold = self.crawler.settings.getint(
            SPIDERMON_MAX_ITEM_VALIDATION_ERRORS, 0
        )
        item_validation_errors = self.stats.get("spidermon/validation/fields/errors", 0)
        msg = "Found {} item validation error. Max allowed is {}.".format(
            item_validation_errors, errors_threshold
        )
        self.assertTrue(item_validation_errors <= errors_threshold, msg=msg)


@monitors.name("Field Coverage Monitor")
class FieldCoverageMonitor(BaseScrapyMonitor):
    """Validate if field coverage rules are met.

    To use this monitor you need to enable the ``SPIDERMON_ADD_FIELD_COVERAGE``
    setting, which will add information about field coverage to your spider
    statistics.

    To define your field coverage rules create a dictionary containing the
    expected coverage for each field you want to monitor.

    As an example, if the items you are returning from your spider are Python dictionaries
    with the following format:

    .. code-block:: python

        {
            "field_1": "some_value",
            "field_2": "some_value",
            "field_3": {
                "field_3_1": "some_value",
                "field_3_2": "some_value",
            }
        }

    A set of rules may be defined as follows:

    .. code-block:: python

        # project/settings.py
        SPIDERMON_FIELD_COVERAGE_RULES = {
            "dict/field_1": 0.4,  # Expected 40% coverage for field_1
            "dict/field_2": 1.0,  # Expected 100% coverage for field_2
            "dict/field_3": 0.8,  # Expected 80% coverage for parent field_3
            "dict/field_3/field_3_1": 0.5,  # Expected 50% coverage for nested field_3_1
        }

    You are not obligated to set rules for every field, just for the ones in which you are interested.
    Also, you can monitor nested fields if available in your returned items.

    .. warning::

       Rules for nested fields will be validated against the total number of items returned.

       For the example below, rule for ``dict/field_3/field_3_1`` will validate if 50%
       of **all** items returned contains ``field_3_1``, not just the ones that contain
       parent ``field_3``.

    .. note::
       If you are returning an item type other than a dictionary, replace `dict` by the
       class name of the item you are returning.

       Considering you have an item defined as:

       .. code-block:: python

           class MyCustomItem(scrapy.Item):
               field_1 = scrapy.Field()
               field_2 = scrapy.Field()

       You must define the field coverage rules as follows:

       .. code-block:: python

           SPIDERMON_FIELD_COVERAGE_RULES = {
               "MyCustomItem/field_1": 0.4,
               "MyCustomItem/field_2": 1.0,
           }"""

    def run(self, result):
        add_field_coverage_set = self.crawler.settings.getbool(
            "SPIDERMON_ADD_FIELD_COVERAGE", False
        )
        if not add_field_coverage_set:
            raise NotConfigured(
                "To enable field coverage monitor, set SPIDERMON_ADD_FIELD_COVERAGE=True in your project settings"
            )
        return super().run(result)

    def test_check_if_field_coverage_rules_are_met(self):
        failures = []
        field_coverage_rules = self.crawler.settings.getdict(
            "SPIDERMON_FIELD_COVERAGE_RULES"
        )
        for field, expected_coverage in field_coverage_rules.items():
            actual_coverage = self.data.stats.get(
                f"spidermon_field_coverage/{field}", 0
            )
            if actual_coverage < expected_coverage:
                failures.append(
                    "{} (expected {}, got {})".format(
                        field, expected_coverage, actual_coverage
                    )
                )

        msg = "\nThe following items did not meet field coverage rules:\n{}".format(
            "\n".join(failures)
        )
        self.assertTrue(len(failures) == 0, msg=msg)


class SpiderCloseMonitorSuite(MonitorSuite):
    """This Monitor Suite implements the following monitors:

    * :class:`ItemCountMonitor`
    * :class:`ItemValidationMonitor`
    * :class:`ErrorCountMonitor`
    * :class:`WarningCountMonitor`
    * :class:`FinishReasonMonitor`
    * :class:`UnwantedHTTPCodesMonitor`
    * :class:`FieldCoverageMonitor`
    * :class:`RetryCountMonitor`
    * :class:`DownloaderExceptionMonitor`
    * :class:`SuccessfulRequestsMonitor`
    * :class:`TotalRequestsMonitor`

    You can easily enable this monitor *after* enabling Spidermon::

            SPIDERMON_SPIDER_CLOSE_MONITORS = (
                'spidermon.contrib.scrapy.monitors.SpiderCloseMonitorSuite',
            )
    """

    monitors = [
        ItemCountMonitor,
        ItemValidationMonitor,
        ErrorCountMonitor,
        WarningCountMonitor,
        FinishReasonMonitor,
        UnwantedHTTPCodesMonitor,
        FieldCoverageMonitor,
        RetryCountMonitor,
        DownloaderExceptionMonitor,
        SuccessfulRequestsMonitor,
        TotalRequestsMonitor,
    ]



import ast
import json
import logging

from slack import WebClient
from slack.errors import SlackApiError

from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured

logger = logging.getLogger(__name__)


class SlackMessageManager:
    sender_token = None
    sender_name = None

    def __init__(self, sender_token=None, sender_name=None, fake=False):
        sender_token = sender_token or self.sender_token
        if not sender_token:
            raise NotConfigured("You must provide a slack token.")

        self.sender_name = sender_name or self.sender_name
        if not self.sender_name:
            raise NotConfigured("You must provide a slack sender name.")

        self.fake = fake
        self._client = WebClient(sender_token)
        self._users = None

    @property
    def users(self):
        if self._users is None:
            self._users = self._get_users_info()
        return self._users

    def send_message(
        self, to, text, parse=None, link_names=1, attachments=None, use_mention=False
    ):
        if self.fake:
            logger.info(text)
            if attachments:
                logger.info(attachments)
            return

        if isinstance(to, list):
            return [
                self.send_message(
                    to=recipient,
                    text=text,
                    parse=parse,
                    link_names=link_names,
                    attachments=attachments,
                    use_mention=use_mention,
                )
                for recipient in to
            ]
        elif to.startswith("@"):
            return self._send_user_message(
                username=to,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
            )
        else:
            if use_mention:
                if to.startswith("#"):
                    text = "@channel: " + text
                else:
                    text = "@group: " + text
            return self._send_channel_message(
                channel=to,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
            )

    def _get_user_id(self, username):
        name = username[1:] if username.startswith("@") else username
        user = self.users.get(name, None)
        return user["id"] if user else None

    def _get_users_info(self):
        return dict(
            [
                (member["name"].lower(), member)
                for member in self._client.users_list()["members"]
            ]
        )

    def _send_user_message(
        self, username, text, parse="full", link_names=1, attachments=None
    ):
        user_id = self._get_user_id(username)
        if user_id:
            return self._send_channel_message(
                channel=user_id,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
            )

    def _send_channel_message(
        self, channel, text, parse="full", link_names=1, attachments=None
    ):
        self._client.chat_postMessage(
            channel=channel,
            text=text,
            parse=parse,
            link_names=link_names,
            attachments=self._parse_attachments(attachments),
            username=self.sender_name,
            icon_url=self._get_icon_url(),
        )

    def _get_icon_url(self):
        """
        Looks up the icon url for the user set as the message sender

        This will only return a URL if the slack app has users:read permission and
        bot appears in the organisation user list. A no frills bot intended to
        just send messages is not likely to fulfill these criteria, so it returns
        None in this situation, which will result in slack using the bot's App Icon.
        """
        try:
            icon_url = self.users[self.sender_name]["profile"]["image_48"]
        except SlackApiError as e:
            if (
                e.response.data.get("error") == "missing_scope"
                and e.response.data.get("needed") == "users:read"
            ):
                logger.warning(
                    f"bot does not have users:read permissions for slack org - default icon url used"
                )
                # can be an expected outcome - will use its own icon
                icon_url = None
            else:
                raise e
        except KeyError:
            # bot has read permissions for slack org but can't find sender in list
            # can be an expected outcome - will use its own icon
            logger.warning(
                f"bot cannot finder user in slack org member list - default icon url used"
            )
            icon_url = None
        return icon_url

    def _parse_attachments(self, attachments):
        if not attachments:
            return None
        else:
            python_attachments = ast.literal_eval(attachments)
            return json.dumps(python_attachments)


class SendSlackMessage(ActionWithTemplates):
    message = None
    attachments = None
    message_template = "slack/default/message.jinja"
    attachments_template = "slack/default/attachments.jinja"
    recipients = None
    sender_token = None
    sender_name = None
    include_message = True
    include_attachments = True
    fake = False

    def __init__(
        self,
        sender_token=None,
        sender_name=None,
        recipients=None,
        message=None,
        message_template=None,
        include_message=None,
        attachments=None,
        attachments_template=None,
        include_attachments=None,
        fake=None,
    ):
        super().__init__()

        self.fake = fake or self.fake
        self.manager = SlackMessageManager(
            sender_token=sender_token or self.sender_token,
            sender_name=sender_name or self.sender_name,
            fake=self.fake,
        )

        self.recipients = recipients or self.recipients
        self.message = message or self.message
        self.message_template = message_template or self.message_template
        self.include_message = include_message or self.include_message
        self.attachments = attachments or self.attachments
        self.attachments_template = attachments_template or self.attachments_template
        self.include_attachments = include_attachments or self.include_attachments
        if not self.fake and not self.recipients:
            raise NotConfigured(
                "You must provide at least one recipient for the message."
            )

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "sender_token": crawler.settings.get("SPIDERMON_SLACK_SENDER_TOKEN"),
            "sender_name": crawler.settings.get("SPIDERMON_SLACK_SENDER_NAME"),
            "recipients": crawler.settings.getlist("SPIDERMON_SLACK_RECIPIENTS"),
            "message": crawler.settings.get("SPIDERMON_SLACK_MESSAGE"),
            "message_template": crawler.settings.get(
                "SPIDERMON_SLACK_MESSAGE_TEMPLATE"
            ),
            "attachments": crawler.settings.get("SPIDERMON_SLACK_ATTACHMENTS"),
            "attachments_template": crawler.settings.get(
                "SPIDERMON_SLACK_ATTACHMENTS_TEMPLATE"
            ),
            "include_message": crawler.settings.getbool(
                "SPIDERMON_SLACK_INCLUDE_MESSAGE"
            ),
            "include_attachments": crawler.settings.getbool(
                "SPIDERMON_SLACK_INCLUDE_ATTACHMENTS"
            ),
            "fake": crawler.settings.getbool("SPIDERMON_SLACK_FAKE"),
        }

    def run_action(self):
        self.manager.send_message(
            to=self.recipients,
            text=self.get_message(),
            attachments=self.get_attachments(),
        )

    def get_message(self):
        if self.include_message:
            if self.message:
                return self.render_text_template(self.message)
            else:
                return self.render_template(self.message_template)
        else:
            return None

    def get_attachments(self):
        if self.include_attachments:
            if self.attachments:
                return self.render_text_template(self.attachments)
            else:
                return self.render_template(self.attachments_template)
        else:
            return None



from spidermon.contrib.monitors.mixins import ValidationMonitorMixin, SpiderMonitorMixin

from .factory import PythonExpressionsMonitor
from .context import Context


class ExpressionsMonitor(
    PythonExpressionsMonitor, ValidationMonitorMixin, SpiderMonitorMixin
):
    def get_context_data(self):
        context = Context()
        attrs = ["stats", "crawler", "spider", "job", "validation", "responses"]
        context.extend_via_attrs(self, attrs)
        return context



UNEXPECTED_FIELD = "Unexpected field"
UNEXPECTED_FIELDS = "Unexpected fields: {unexpected_fields}"
MISSING_REQUIRED_FIELD = "Missing required field"
VALUE_NOT_IN_CHOICES = "Value not in choices"
FIELD_TOO_LONG = "Field too long"
FIELD_TOO_SHORT = "Field too short"
INVALID_STRING = "Invalid string"
REGEX_NOT_MATCHED = "Regex not matched"
INVALID_DATETIME = "Invalid datetime"
INVALID_DATE = "Invalid date"
INVALID_INT = "Invalid int"
INVALID_FLOAT = "Invalid float"
INVALID_LONG = "Invalid long"
INVALID_DECIMAL = "Invalid decimal"
INVALID_BOOLEAN = "Invalid boolean"
NUMBER_TOO_HIGH = "Number too high"
NUMBER_TOO_LOW = "Number too low"
INVALID_EMAIL = "Invalid email"
INVALID_URL = "Invalid URL"
INVALID_UUID = "Invalid UUID"
INVALID_IPV4 = "Invalid IPv4"
INVALID_IPV6 = "Invalid IPv6"
INVALID_HOSTNAME = "Invalid hostname"
INVALID_URI = "Invalid URI"
INVALID_REGEX = "Invalid regex"
INVALID_COLOR = "Invalid color"
INVALID_HASH_LENGTH = "Invalid hash length"
INVALID_HASH = "Invalid hash"
LIST_TOO_SHORT = "List too short"
LIST_TOO_LONG = "List too long"
INVALID_LIST = "Invalid list"
INVALID_DICT = "Invalid dict"
INVALID_CHILD_CONTENT = "Invalid child content"
INVALID_NUMBER = "Invalid number"
INVALID_OBJECT = "Invalid object"
INVALID_ARRAY = "Invalid array"
NOT_NULL = "Not null"
NOT_VALID_UNDER_ANY_SCHEMA = "Not valid under any schema"
VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS = "Valid for several exclusive schemas"
TOO_MANY_ITEMS = "Too many items"
MISSING_DEPENDENT_FIELD = "Missing dependent field"
TOO_MANY_PROPERTIES = "Too many properties"
NOT_ENOUGH_PROPERTIES = "Not enough properties"
NOT_MULTIPLE_OF = "Not multiple of"
NOT_ALLOWED_VALUE = "Not allowed value"
NOT_UNIQUE = "Not unique"



from . import SendTelegramMessage


class SendTelegramMessageSpiderStarted(SendTelegramMessage):
    message_template = "telegram/spider/notifier/start/message.jinja"


class SendTelegramMessageSpiderFinished(SendTelegramMessage):
    message_template = "telegram/spider/notifier/finish/message.jinja"
    include_ok_messages = False
    include_error_messages = True

    def __init__(
        self, include_ok_messages=None, include_error_messages=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.include_ok_messages = include_ok_messages or self.include_ok_messages
        self.include_error_messages = (
            include_error_messages or self.include_error_messages
        )

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        kwargs.update(
            {
                "include_ok_messages": crawler.settings.get(
                    "SPIDERMON_TELEGRAM_NOTIFIER_INCLUDE_OK_MESSAGES"
                ),
                "include_error_messages": crawler.settings.get(
                    "SPIDERMON_TELEGRAM_NOTIFIER_INCLUDE_ERROR_MESSAGES"
                ),
            }
        )
        return kwargs

    def get_template_context(self):
        context = super().get_template_context()
        context.update(
            {
                "include_ok_messages": self.include_ok_messages,
                "include_error_messages": self.include_error_messages,
            }
        )
        return context


class SendTelegramMessageSpiderRunning(SendTelegramMessageSpiderFinished):
    message_template = "telegram/spider/notifier/periodic/message.jinja"



from spidermon.exceptions import NotConfigured

from . import CreateReport


class CreateFileReport(CreateReport):
    filename = None

    def __init__(self, filename, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = filename or self.filename
        if not self.filename:
            raise NotConfigured("You must define a template output file.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        kwargs.update({"filename": crawler.settings.get("SPIDERMON_REPORT_FILENAME")})
        return kwargs

    def after_render_report(self):
        rendered_filename = self.render_text_template(self.filename)

        with open(rendered_filename, "w") as f:
            f.write(self.report)



import schematics
from schematics.exceptions import ValidationError
from schematics.models import Model
from schematics.types import (
    StringType,
    DateTimeType,
    DateType,
    FloatType,
    IntType,
    LongType,
    DecimalType,
    BooleanType,
    EmailType,
    URLType,
    UUIDType,
    IPv4Type,
    MD5Type,
    SHA1Type,
)
from schematics.types.compound import ListType, DictType, ModelType

from spidermon.contrib.validation import SchematicsValidator, messages


SCHEMATICS1 = schematics.__version__.startswith("1.")


def test_rogue_fields():
    """
    messages:
        - UNEXPECTED_FIELD
    """
    _test_data(
        model=Model, data={"a": 1}, expected=(False, {"a": [messages.UNEXPECTED_FIELD]})
    )
    _test_data(model=Model, data={"a": 1}, expected=(True, {}), strict=False)


def test_required():
    """
    messages:
        - MISSING_REQUIRED_FIELD
    """

    class DataRequired(Model):
        a = StringType(required=True)

    class DataNotRequired(Model):
        a = StringType(required=False)

    _test_data(
        model=DataRequired,
        data={},
        expected=(False, {"a": [messages.MISSING_REQUIRED_FIELD]}),
    )
    _test_data(model=DataNotRequired, data={}, expected=(True, {}))


def test_choices():
    """
    messages:
        - VALUE_NOT_IN_CHOICES
    """

    class Data(Model):
        a = StringType(choices=["a", "b"])
        b = IntType(choices=[1, 2, 3])

    _test_data(model=Data, data={}, expected=(True, {}))
    _test_data(model=Data, data={"a": "b", "b": 3}, expected=(True, {}))
    _test_data(
        model=Data,
        data={"a": "c", "b": 4},
        expected=(
            False,
            {
                "a": [messages.VALUE_NOT_IN_CHOICES],
                "b": [messages.VALUE_NOT_IN_CHOICES],
            },
        ),
    )


def test_string_valid():
    """
    messages:
        - INVALID_STRING
    """

    class Data(Model):
        a = StringType()

    _test_data(model=Data, data={"a": "hello there!"}, expected=(True, {}))
    _test_data(
        model=Data, data={"a": []}, expected=(False, {"a": [messages.INVALID_STRING]})
    )


def test_string_lengths():
    """
    messages:
        - FIELD_TOO_SHORT
        - FIELD_TOO_LONG
    """

    class Data(Model):
        a = StringType(min_length=2, max_length=5)

    _test_data(model=Data, data={"a": "12"}, expected=(True, {}))
    _test_data(model=Data, data={"a": "12345"}, expected=(True, {}))
    _test_data(
        model=Data, data={"a": "1"}, expected=(False, {"a": [messages.FIELD_TOO_SHORT]})
    )
    _test_data(
        model=Data,
        data={"a": "123456"},
        expected=(False, {"a": [messages.FIELD_TOO_LONG]}),
    )


def test_string_regex():
    """
    messages:
        - REGEX_NOT_MATCHED
    """

    class Data(Model):
        a = StringType(regex=".*def.*")

    _test_data(
        model=Data,
        data={"a": "abc"},
        expected=(False, {"a": [messages.REGEX_NOT_MATCHED]}),
    )
    _test_data(model=Data, data={"a": "abcdefghi"}, expected=(True, {}))
    _test_data(model=Data, data={"a": "def"}, expected=(True, {}))


def test_datetime():
    """
    messages:
        - INVALID_DATETIME
    """

    class Data(Model):
        a = DateTimeType()

    class DataWithFormats(Model):
        a = DateTimeType(formats=("%Y-%m-%d %H:%M:%S.%f", "%Y-%m-%d %H:%M:%S"))

    if SCHEMATICS1:
        INVALID = ["2015-05-13 13:35:15.718978", "2015-05-13 13:35:15"]
    else:
        INVALID = ["foo", "1-2-3"]
        CUSTOM_FORMAT = ["2015-05-13 13:35:15.718978", "2015-05-13 13:35:15"]
    VALID = ["2015-05-13T13:35:15.718978", "2015-05-13T13:35:15"]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DATETIME,
    )
    if SCHEMATICS1:
        for dt in INVALID:
            _test_data(model=DataWithFormats, data={"a": dt}, expected=(True, {}))
    else:
        for dt in CUSTOM_FORMAT:
            _test_data(model=DataWithFormats, data={"a": dt}, expected=(True, {}))


def test_date():
    """
    messages:
        - INVALID_DATE
    """

    class Data(Model):
        a = DateType()

    INVALID = ["2015-05-13 13:35:15", "13-05-2013", "2015-20-13", "2015-01-40"]
    VALID = ["2015-05-13", "2050-01-01"]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_DATE
    )


def test_int():
    """
    messages:
        - INVALID_INT
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = IntType(min_value=-10, max_value=10)
        b = IntType()

    INVALID = ["", "a", "2a", "2015-05-13 13:35:15", "7.2"]
    VALID = ["1", "8", "-2", "-7", 1, 8, -2, -7]
    if SCHEMATICS1:
        VALID.append(7.2)
    else:
        INVALID.append(7.2)
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_INT
    )
    _test_data(
        model=Data, data={"a": -20}, expected=(False, {"a": [messages.NUMBER_TOO_LOW]})
    )
    _test_data(
        model=Data, data={"a": 11}, expected=(False, {"a": [messages.NUMBER_TOO_HIGH]})
    )


def test_float():
    """
    messages:
        - INVALID_FLOAT
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = FloatType(min_value=-10, max_value=10)

    INVALID = ["", "a", "2a", "2015-05-13 13:35:15"]
    VALID = [
        "1",
        "-2",
        "8",
        "2.3",
        "5.2354958",
        "-9.231",
        1,
        -2,
        8,
        2.3,
        5.2354958,
        -9.231,
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_FLOAT
    )
    _test_data(
        model=Data, data={"a": -20}, expected=(False, {"a": [messages.NUMBER_TOO_LOW]})
    )
    _test_data(
        model=Data, data={"a": 11}, expected=(False, {"a": [messages.NUMBER_TOO_HIGH]})
    )


def test_long():
    """
    messages:
        - INVALID_LONG
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = LongType(min_value=-10, max_value=10)

    INVALID = ["", "a", "2a", "2015-05-13 13:35:15", "2.3", "5.2354958"]
    VALID = ["1", "-2", "8", 1, -2, 8]
    if SCHEMATICS1:
        expected_error = messages.INVALID_LONG
    else:
        expected_error = messages.INVALID_INT
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=expected_error
    )
    _test_data(
        model=Data, data={"a": -20}, expected=(False, {"a": [messages.NUMBER_TOO_LOW]})
    )
    _test_data(
        model=Data, data={"a": 11}, expected=(False, {"a": [messages.NUMBER_TOO_HIGH]})
    )


def test_decimal():
    """
    messages:
        - INVALID_DECIMAL
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = DecimalType(min_value=-10, max_value=10)

    INVALID = ["", "a", "2a", "2015-05-13 13:35:15"]
    VALID = [
        "1",
        "-2",
        "8",
        "2.3",
        "5.2354958",
        "-9.231",
        1,
        -2,
        8,
        2.3,
        5.2354958,
        -9.231,
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DECIMAL,
    )
    _test_data(
        model=Data, data={"a": -20}, expected=(False, {"a": [messages.NUMBER_TOO_LOW]})
    )
    _test_data(
        model=Data, data={"a": 11}, expected=(False, {"a": [messages.NUMBER_TOO_HIGH]})
    )


def test_boolean():
    """
    messages:
        - INVALID_BOOLEAN
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = BooleanType()

    INVALID = ["", "a", "2" "TRUE", "FALSE", "TruE", "FalsE"]
    VALID = [0, 1, "0", "1", "True", "False", "true", "false", True, False]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_BOOLEAN,
    )


def test_email():
    """
    messages:
        - INVALID_EMAIL
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """

    class Data(Model):
        a = EmailType()

    INVALID = [
        "",
        "johndoe",
        "johndoe@domain" "johndoe@domain." "@domain" "@domain.com" "domain.com",
    ]
    VALID = [
        "johndoe@domain.com",
        "john.doe@domain.com",
        "john.doe@sub.domain.com",
        "j@sub.domain.com",
        "j@d.com",
        "j@domain.co.uk",
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_EMAIL
    )


def test_url():
    """
    messages:
        - INVALID_URL
    """

    class Data(Model):
        a = URLType()

    INVALID = [
        "",
        "http://",
        "http://www.",
        "www.",
        "http://www. .com",
        "domain.com",
        "www.domain.com",
        "http:/www.domain.com",
        "http//www.domain.com",
        "http:www.domain.com",
        "htp://domain.com/",
        "http://sub.domain.com\\en-us\\default.aspx\\",
        "http:\\\\msdn.domain.com\\en-us\\library\\default.aspx\\",
        "http:\\\\www.domain.com\\leafnode-L1.html",
        "./",
        "../",
        "http:\\\\www.domain.com\\leafnode-L1.xhtml\\",
    ]
    VALID = [
        "http://www.domain",
        "http://www.com",
        "http://www.domain.com.",
        "http://www.domain.com/.",
        "http://www.domain.com/..",
        "http://www.domain.com//cataglog//index.html",
        "http://www.domain.net/",
        "http://www.domain.com/level2/leafnode-L2.xhtml/",
        "http://www.domain.com/level2/level3/leafnode-L3.xhtml/",
        "http://www.domain.com?pageid=123&testid=1524",
        "http://www.domain.com/do.html#A",
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_URL
    )


def test_uuid():
    """
    messages:
        - INVALID_UUID
    """

    class Data(Model):
        a = UUIDType()

    INVALID = [
        "",
        "678as6sd88ads67",
        "678as6sd88ads67-alskjlasd",
        "xxxxxxxx-xxxx-Mxxx-Nxxx-xxxxxxxxxxxx",
        "2.25.290383009913173870543740933812899923227",
    ]
    VALID = [
        "12345678-1234-5678-1234-567812345678",
        "12345678123456781234567812345678",
        "urn:uuid:12345678-1234-5678-1234-567812345678",
        "cfc63f3f-f3a7-465a-8183-acf055c6d472",
        "00000000-0000-0000-0000-000000000000",
        "01234567-89ab-cdef-0123456789abcdef",
        "0123456789abcdef0123456789abcdef",
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_UUID
    )


def test_ipv4type():
    """
    messages:
        - INVALID_IPV4
    """

    class Data(Model):
        a = IPv4Type()

    INVALID = [
        "",
        "0",
        "0.",
        "0.0",
        "0.0.",
        "0.0.0",
        "0.0.0.0.",
        "0.0.0.0.0",
        "256.256.256.256",
        "2002:4559:1FE2::4559:1FE2",
        "2002:4559:1FE2:0:0:0:4559:1FE2",
        "2002:4559:1FE2:0000:0000:0000:4559:1FE2",
    ]
    VALID = [
        "98.139.180.149",
        "69.89.31.226",
        "192.168.1.1",
        "127.0.0.0",
        "0.0.0.0",
        "255.255.255.255",
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_IPV4
    )


def test_md5():
    """
    messages:
        - INVALID_HASH
        - INVALID_HASH_LENGTH
    """

    class Data(Model):
        a = MD5Type()

    INVALID = [
        "_b1a9953c4611296a827abf8a47804d7",
        "zb1a9953c4611296a827abf8a47804d7",
        "Gb1a9953c4611296a827abf8a47804d7",
        # FIXME: PY3: schematics uses integer conversion for validating hex and
        # Py3 integers can contain underscores.
        # '8b1_9953c4611296a827abf8c47804d1',
    ]
    VALID = [
        "8b1a9953c4611296a827abf8c47804d7",
        "7dd4bbe8a38600b556f79ca44c9b5132",
        "11111111111111111111111111111111",
        "8B1A9953C4611296A827ABF8C47804D1",
    ]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_HASH
    )
    _test_data(
        model=Data,
        data={"a": "8b1a9953c4611296a827abf8c47804d"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={"a": "8b1a9953c4611296"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={"a": "8b1a9953c46112968b1a9953c46112968b1a9953c4611296"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )


def test_sha1():
    """
    messages:
        - INVALID_HASH
        - INVALID_HASH_LENGTH
    """

    class Data(Model):
        a = SHA1Type()

    INVALID = [
        "_03d40e1a2ede7e31f3c3b45a9e87d12ed33402e",
        "g03d40e1a2ede7e31f3c3b45a9e87d12ed33402e",
        "z03d40e1a2ede7e31f3c3b45a9e87d12ed33402e",
        "G03d40e1a2ede7e31f3c3b45a9e87d12ed33402e",
        # FIXME: PY3: schematics uses integer conversion for validating hex and
        # Py3 integers can contain underscores.
        # 'a03d_0e1a2ede7e31f3c3b45a9e87d12ed33402e',
    ]
    VALID = ["a03d70e1a2ede7e31f3c3b45a9e87d12ed33402e"]
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_HASH
    )
    _test_data(
        model=Data,
        data={"a": "a03d70e1a2ede7e31f3c3b45a9e87d12ed33402"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={"a": "8b1a9953c4611296"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={"a": "8b1a9953c46112968b1a9953c46112968b1a9953c4611296"},
        expected=(False, {"a": [messages.INVALID_HASH_LENGTH]}),
    )


def test_list():
    """
    messages:
        - INVALID_LIST
        - LIST_TOO_SHORT
        - LIST_TOO_LARGE
        - INVALID_INT
    """

    class Data(Model):
        a = ListType(field=IntType(), min_size=3, max_size=5)

    _test_data(model=Data, data={"a": [1, 2, 3]}, expected=(True, {}))
    _test_data(model=Data, data={"a": ["1", "2", "3"]}, expected=(True, {}))
    _test_data(
        model=Data, data={"a": Data}, expected=(False, {"a": [messages.INVALID_LIST]})
    )
    if SCHEMATICS1:
        _test_data(
            model=Data,
            data={"a": ["a", "b", "c"]},
            expected=(False, {"a": [messages.INVALID_INT]}),
        )
    else:
        _test_data(
            model=Data,
            data={"a": ["a", "b", "c"]},
            expected=(
                False,
                {
                    "a.0": [messages.INVALID_INT],
                    "a.1": [messages.INVALID_INT],
                    "a.2": [messages.INVALID_INT],
                },
            ),
        )
    _test_data(
        model=Data,
        data={"a": [1, 2]},
        expected=(False, {"a": [messages.LIST_TOO_SHORT]}),
    )
    _test_data(
        model=Data,
        data={"a": [1, 2, 3, 4, 5, 6]},
        expected=(False, {"a": [messages.LIST_TOO_LONG]}),
    )


def test_dict():
    """
    messages:
        - INVALID_DICT
        - INVALID_INT
    """

    class Data(Model):
        a = DictType(field=IntType)

    INVALID = ["a", Data]
    VALID = [{}, {"some": 1}]
    if SCHEMATICS1:
        VALID.append([])
    else:
        INVALID.append([])
    _test_valid_invalid(
        model=Data, valid=VALID, invalid=INVALID, expected_error=messages.INVALID_DICT
    )
    if SCHEMATICS1:
        _test_data(
            model=Data,
            data={"a": {"some": "a"}},
            expected=(False, {"a": [messages.INVALID_INT]}),
        )
    else:
        _test_data(
            model=Data,
            data={"a": {"some": "a"}},
            expected=(False, {"a.some": [messages.INVALID_INT]}),
        )


def test_models():
    """
    messages:
        - UNEXPECTED_FIELD
        - MISSING_REQUIRED_FIELD
        - INVALID_FLOAT
    """

    class Coordinates(Model):
        latitude = FloatType(required=True)
        longitude = FloatType(required=True)

    class Geo(Model):
        coordinates = ModelType(Coordinates, required=True)

    class Data(Model):
        geo = ModelType(Geo, required=True)

    _test_data(
        model=Data,
        data={"a": {}},
        expected=(
            False,
            {
                "a": [messages.UNEXPECTED_FIELD],
                "geo": [messages.MISSING_REQUIRED_FIELD],
            },
        ),
    )
    _test_data(
        model=Data,
        data={"geo": None},
        expected=(False, {"geo": [messages.MISSING_REQUIRED_FIELD]}),
    )
    _test_data(
        model=Data,
        data={"geo": {}},
        expected=(False, {"geo.coordinates": [messages.MISSING_REQUIRED_FIELD]}),
    )
    _test_data(
        model=Data,
        data={"geo": {"coordinates": None}},
        expected=(False, {"geo.coordinates": [messages.MISSING_REQUIRED_FIELD]}),
    )
    _test_data(
        model=Data,
        data={"geo": {"coordinates": {}}},
        expected=(
            False,
            {
                "geo.coordinates.latitude": [messages.MISSING_REQUIRED_FIELD],
                "geo.coordinates.longitude": [messages.MISSING_REQUIRED_FIELD],
            },
        ),
    )
    _test_data(
        model=Data,
        data={"geo": {"coordinates": {"latitude": None, "longitude": None}}},
        expected=(
            False,
            {
                "geo.coordinates.latitude": [messages.MISSING_REQUIRED_FIELD],
                "geo.coordinates.longitude": [messages.MISSING_REQUIRED_FIELD],
            },
        ),
    )
    _test_data(
        model=Data,
        data={"geo": {"coordinates": {"latitude": "y", "longitude": "x"}}},
        expected=(
            False,
            {
                "geo.coordinates.latitude": [messages.INVALID_FLOAT],
                "geo.coordinates.longitude": [messages.INVALID_FLOAT],
            },
        ),
    )
    _test_data(
        model=Data,
        data={"geo": {"coordinates": {"latitude": 40.42, "longitude": -3.71}}},
        expected=(True, {}),
    )


def test_multiple_errors_per_field():
    """
    messages:
        - FIELD_TOO_SHORT
        - REGEX_NOT_MATCHED
    """

    class Data(Model):
        a = StringType(min_length=3, regex=r"foo")

    data = {"a": "z"}
    v = SchematicsValidator(Data)
    result = v.validate(data, strict=True)
    assert result[0] is False
    error_messages = result[1]
    assert "a" in error_messages
    expected = [messages.FIELD_TOO_SHORT, messages.REGEX_NOT_MATCHED]
    assert sorted(error_messages["a"]) == expected


def _test_data(model, data, expected, strict=True):
    v = SchematicsValidator(model)
    assert expected == v.validate(data, strict=strict)


def _test_valid_invalid(model, valid, invalid, expected_error, expected_field="a"):
    for dt in valid:
        _test_data(model=model, data={expected_field: dt}, expected=(True, {}))
    for dt in invalid:
        _test_data(
            model=model,
            data={expected_field: dt},
            expected=(False, {expected_field: [expected_error]}),
        )


def test_validation_error_on_model_level_validation():
    class TestModel(Model):
        field_a = StringType()

        def validate_field_a(self, data, value):
            raise ValidationError("Model-level validation failed.")

    _test_data(
        model=TestModel,
        data={"field_a": "some_data"},
        expected=(False, {"field_a": ["Model-level validation failed."]}),
    )



import pytest

from spidermon import MonitorSuite
from spidermon.exceptions import (
    InvalidMonitor,
    InvalidMonitorIterable,
    InvalidMonitorClass,
    InvalidMonitorTuple,
    NotAllowedMethod,
)

from .fixtures.suites import *


class SuiteDefinition:
    def __init__(self, suite_class, monitors=None, expected_number_of_monitors=0):
        self.suite_class = suite_class
        self.monitors = monitors or []
        self.expected_number_of_monitors = expected_number_of_monitors


CLASS_SUITE_DEFINITIONS = [
    # ---------------------------------------------------------------------------------
    #               suite_class    monitors               expected_number_of_monitors
    # ---------------------------------------------------------------------------------
    # no monitors
    SuiteDefinition(EmptySuite, [], 0),
    SuiteDefinition(Suite01, [], 3),
    SuiteDefinition(Suite02, [], 5),
    SuiteDefinition(Suite03, [], 8),
    SuiteDefinition(Suite04, [], 13),
    # monitors as classes
    SuiteDefinition(EmptySuite, [EmptyMonitor], 0),
    SuiteDefinition(EmptySuite, [Monitor01], 3),
    SuiteDefinition(EmptySuite, [Monitor02], 2),
    SuiteDefinition(EmptySuite, [Monitor01, Monitor02], 5),
    SuiteDefinition(EmptySuite, [Suite01], 3),
    SuiteDefinition(EmptySuite, [Suite01, Suite02], 8),
    SuiteDefinition(EmptySuite, [Suite01, Suite02, Monitor01, Monitor02], 13),
    SuiteDefinition(Suite01, [Suite01, Suite02, Monitor01, Monitor02], 16),
    SuiteDefinition(Suite02, [Suite01, Suite02, Monitor01, Monitor02], 18),
    SuiteDefinition(Suite03, [Suite01, Suite02, Monitor01, Monitor02], 21),
    SuiteDefinition(Suite04, [Suite01, Suite02, Monitor01, Monitor02], 26),
    # monitors as instances
    SuiteDefinition(EmptySuite, [Suite01()], 3),
    SuiteDefinition(EmptySuite, [Suite02()], 5),
    SuiteDefinition(EmptySuite, [Suite03()], 8),
    SuiteDefinition(EmptySuite, [Suite04()], 13),
    SuiteDefinition(EmptySuite, [Suite01(), Suite02(), Suite03(), Suite04()], 29),
    # monitors as tuples
    SuiteDefinition(EmptySuite, [("M1", Monitor01)], 3),
    SuiteDefinition(EmptySuite, [("M1", Monitor01), ("M2", Monitor02)], 5),
    SuiteDefinition(EmptySuite, [("S", Suite04)], 13),
]
INVALID_SUITE_DEFINITION_NON_ITERABLE = SuiteDefinition(EmptySuite, 10)
INVALID_SUITE_DEFINITION_NONE = SuiteDefinition(EmptySuite, [None])
INVALID_SUITE_DEFINITION_NUMBER = SuiteDefinition(EmptySuite, [10])
INVALID_SUITE_DEFINITION_CLASS = SuiteDefinition(EmptySuite, [object])
INVALID_SUITE_DEFINITION_TUPLE1 = SuiteDefinition(EmptySuite, [(1,)])
INVALID_SUITE_DEFINITION_TUPLE3 = SuiteDefinition(EmptySuite, [(1, 2, 3)])
INVALID_SUITE_DEFINITION_TUPLE2_WRONG1 = SuiteDefinition(EmptySuite, [(1, 2)])
INVALID_SUITE_DEFINITION_TUPLE2_WRONG2 = SuiteDefinition(EmptySuite, [("A", 2)])

INVALID_SUITE_DEFINITIONS = [
    # --------------------------------------------------------------
    # definition                                raises
    # --------------------------------------------------------------
    (INVALID_SUITE_DEFINITION_NONE, InvalidMonitor),
    (INVALID_SUITE_DEFINITION_NUMBER, InvalidMonitor),
    (INVALID_SUITE_DEFINITION_CLASS, InvalidMonitorClass),
    (INVALID_SUITE_DEFINITION_TUPLE1, InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE3, InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE2_WRONG1, InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE2_WRONG2, InvalidMonitor),
]


def test_creation_error_non_iterable():
    with pytest.raises(InvalidMonitorIterable):
        _test_creation_from_init(INVALID_SUITE_DEFINITION_NON_ITERABLE)
    with pytest.raises(InvalidMonitorIterable):
        _test_creation_from_add_monitors(INVALID_SUITE_DEFINITION_NON_ITERABLE)


def test_creation_error_invalid_from_init():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_init(definition)


def test_creation_error_invalid_from_add_monitors():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_add_monitors(definition)


def test_creation_error_invalid_from_add_monitor():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_add_monitor(definition)


def test_creation_from_init():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_init(definition)


def test_creation_from_add_monitors():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_add_monitors(definition)


def test_creation_from_add_monitor():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_add_monitor(definition)


def _test_creation_from_init(definition):
    suite = definition.suite_class(monitors=definition.monitors)
    check_suite(
        suite=suite, expected_number_of_monitors=definition.expected_number_of_monitors
    )


def test_not_allowed_methods():
    suite = EmptySuite()
    with pytest.raises(NotAllowedMethod):
        suite.addTest()
    with pytest.raises(NotAllowedMethod):
        suite.addTests()


def _test_creation_from_add_monitors(definition):
    suite = definition.suite_class()
    suite.add_monitors(definition.monitors)
    check_suite(
        suite=suite, expected_number_of_monitors=definition.expected_number_of_monitors
    )


def _test_creation_from_add_monitor(definition):
    suite = definition.suite_class()
    for monitor in definition.monitors:
        suite.add_monitor(monitor)
    check_suite(
        suite=suite, expected_number_of_monitors=definition.expected_number_of_monitors
    )


def check_suite(suite, expected_number_of_monitors):
    # print
    # print suite.debug_tree()
    # print
    assert isinstance(suite, MonitorSuite)
    assert suite.number_of_monitors == expected_number_of_monitors
    for monitor in suite:
        assert isinstance(monitor, (Monitor, MonitorSuite))
    all_monitors = suite.all_monitors
    assert len(all_monitors) == expected_number_of_monitors
    for test in all_monitors:
        assert isinstance(test, (Monitor, Monitor))



import re


def calculate_field_coverage(stats):
    coverage = {}
    for key, value in stats.items():
        if not key.startswith("spidermon_item_scraped_count"):
            continue

        item_type_m = re.search(
            r"spidermon_item_scraped_count\/(?P<item_type>\w+)\/(?P<item_key>.*)", key
        )
        if item_type_m:
            item_type = item_type_m.group(1)
            item_key = item_type_m.group(2)

            item_type_total = stats.get(f"spidermon_item_scraped_count/{item_type}")
            field_coverage = value / item_type_total

            coverage[
                f"spidermon_field_coverage/{item_type}/{item_key}"
            ] = field_coverage

    return coverage



DEFAULT_LINE_LENGTH = 70
DEFAULT_LINE_CHAR = "-"
DEFAULT_LINE_BOLD_CHAR = "="
DEFAULT_LINE_LIGHT_CHAR = "-"


def line(length=DEFAULT_LINE_LENGTH, char=DEFAULT_LINE_CHAR):
    return char * length


def line_title(title, length=DEFAULT_LINE_LENGTH, char=DEFAULT_LINE_CHAR):
    title_length = len(title) + 2
    left_length = (length - title_length) // 2
    right_length = left_length + length - title_length - left_length * 2
    return "{} {} {}".format(char * left_length, title, char * right_length)


class Message:
    def __init__(self, text=None):
        self.msg = text or ""

    def write(self, text):
        self.msg += text

    def write_line(self, text=None):
        self.msg += "%s\n" % (text or "")

    def write_separator(self, length=DEFAULT_LINE_LENGTH, char=DEFAULT_LINE_CHAR):
        self.write_line(line(length=length, char=char))

    def write_bold_separator(self, length=DEFAULT_LINE_LENGTH):
        self.write_separator(length=length, char=DEFAULT_LINE_BOLD_CHAR)

    def write_light_separator(self, length=DEFAULT_LINE_LENGTH):
        self.write_separator(length=length, char=DEFAULT_LINE_LIGHT_CHAR)

    def __str__(self):
        return self.msg



SYNTAXERROR_EXPRESSIONS = ["?", "a string", "a="]

INVALID_EXPRESSIONS = [
    None,
    0,
    "",
    """
    """,
    "a > b\na > b",
    "import os",
    "from os import *",
    "for i in range(10): pass",
    "def something(): pass",
    "lambda x: x",
    "del a",
    # "print(a)",
    "raise Exception",
    "pass",
    "return",
    "yield x",
    # Assignment Operations
    "a = 1",
    "a += 1",
    "a -= 1",
    "a *= 1",
    "a /= 1",
    "a %= 1",
    "a **= 1",
    "a //= 1",
]

VALID_EXPRESSIONS = [
    # Strings
    "'single quoted string'",
    '"double quoted string"',
    "u'unicode single quoted string'",
    'u"unicode double quoted string"',
    # Numbers
    "0",
    ".5",
    "-10",
    "10",
    "1.0",
    "-1.0",
    "3.14j",
    # Sequences
    "[1, 2, 3]",
    "{'a': 1, 'b': 1}",
    "set()",
    "(1, 2)",
    "len(a)",
    # Constants
    "None",
    "True",
    "False",
    # Arithmetic Operations
    "-a",
    "a + 1",
    "a - 1",
    "a / 1",
    "a * 1",
    "a % 1",
    "a ** 1",
    "a // 1",
    # Comparison Operations
    "a == b",
    "a != b",
    "a > b",
    "a < b",
    "a >= b",
    "a <= b",
    "1 < a < 10",
    # Bitwise Operations
    "a & b",
    "a | b",
    "a ^ b",
    "~a",
    "a << b",
    "a >> b",
    # Logical Operations
    "a and b",
    "a or b",
    "not a",
    # Membership Operations
    "a in b",
    "a not in b",
    # Identity Operations
    "a is 10",
    "a is not 10",
    # Inline if statement
    "a if b else c",
    # Subscripting
    "a[1]",
    "a[1:2]",
    "a[1:2, 3]",
    # Comprehensions
    "[i for i in range(10)]",
    "{i: i**2 for i in range(10)}",
    "{i for i in range(10)}",
    "[n for n in range(10) if n>5]",
    # Attribute access
    "stats.scraped_items",
]


EXPRESSIONS_TO_EVALUATE = [
    ("stats.item_scraped_count == 10000", False),
    ("stats.item_scraped_count != 10000", True),
    ("stats.item_scraped_count < 10000", False),
    ("stats.item_scraped_count > 10000", True),
    ("stats.item_scraped_count <= 10000", False),
    ("stats.item_scraped_count >= 10000", True),
    ("10000 < stats.item_scraped_count < 100000", True),
    ("stats.item_scraped_count == .5", False),
    ("stats.item_scraped_count < .5", False),
    ("stats.item_scraped_count > .5", True),
    ('stats.finish_reason in ["finished"]', True),
    ('stats.finish_reason in ["a", "b"]', False),
    ('stats.finish_reason not in ["a", "b"]', True),
    ('stats.finish_reason in {"finished": 0}', True),
    ('stats.finish_reason in {"finished": 0}.keys()', True),
    ('stats.finish_reason in ("finished",)', True),
    ('len(stats.finish_reason) == len("finished")', True),
    ("stats.finish_reason is not None", True),
    ("len(stats.finish_reason) > 0", True),
    ("(len(stats.finish_reason) > 0) == True", True),
    ("(len(stats.finish_reason) > 0) != False", True),
    ("(len(stats.finish_reason) > 0) is not False", True),
    ("(len(stats.finish_reason) > 0) is True", True),
    ('stats["downloader/exception_count"] == 16', True),
    ('-stats["downloader/exception_count"] == -16', True),
    ('stats["downloader/exception_count"]+1 == 17', True),
    ('stats["downloader/exception_count"]-1 == 15', True),
    ('stats["downloader/exception_count"]/2 == 8', True),
    ('stats["downloader/exception_count"]*2 == 32', True),
    ('stats["downloader/exception_count"]%3 == 1', True),
    ('stats["downloader/exception_count"]**2 == 256', True),
    ('stats["downloader/exception_count"]//2 == 8', True),
    ('stats["downloader/exception_count"]*2/2 == 16', True),
    ("stats.has_errors and stats.has_redirections", True),
    ("stats.has_errors or stats.has_redirections", True),
    ("not stats.has_errors", False),
    ("stats.has_errors is True", True),
    ("stats.has_errors is not True", False),
    ("True if stats.has_errors else False", True),
    ("True if not stats.has_errors else False", False),
    ("stats.item_scraped_count in range(29830, 29840)", True),
]



from collections import OrderedDict
import unittest

from spidermon import settings

from .steps import MonitorStep, ActionsStep


def step_required_decorator(allowed_steps):
    def _step_required_decorator(fn):
        def decorator(self, *args, **kwargs):
            if self.step.name not in allowed_steps:
                raise ValueError  # TO-DO
            else:
                return fn(self, *args, **kwargs)

        return decorator

    return _step_required_decorator


monitors_step_required = step_required_decorator(settings.STEPS.MONITOR_RELATED)
actions_step_required = step_required_decorator(settings.STEPS.ACTION_RELATED)


class MonitorResult(unittest.TestResult):
    def __init__(self):
        super().__init__()
        self._steps = OrderedDict(
            [(step, self._get_step_class(step)(step)) for step in settings.STEPS.ALL]
        )
        self._current_step = None

    @property
    def all_monitors_passed(self):
        return self._step_monitors.successful

    @property
    def monitor_results(self):
        return self._step_monitors.all_items

    @property
    def monitors_passed_results(self):
        return self._step_monitors.items_for_statuses(
            settings.MONITOR.STATUSES.SUCCESSFUL
        )

    @property
    def monitors_failed_results(self):
        return self._step_monitors.items_for_statuses(settings.MONITOR.STATUSES.ERROR)

    @property
    def monitors_finished_action_results(self):
        return self._step_monitors_finished.all_items

    @property
    def monitors_passed_action_results(self):
        return self._step_monitors_passed.all_items

    @property
    def monitors_failed_action_results(self):
        return self._step_monitors_failed.all_items

    @property
    def step(self):
        return self._current_step

    def start(self):
        pass

    def finish(self):
        pass

    def next_step(self):
        index = (
            0 if not self.step else list(self._steps.keys()).index(self.step.name) + 1
        )
        self._current_step = list(self._steps.items())[index][1]
        self.step.start()

    def finish_step(self):
        self.step.finish()

    @monitors_step_required
    def startTest(self, test):
        super().startTest(test)
        self.step.add_item(test)

    @monitors_step_required
    def addSuccess(self, test):
        super().addSuccess(test)
        self.step[test].status = settings.MONITOR.STATUS.SUCCESS

    @monitors_step_required
    def addError(self, test, error):
        super().addError(test, error)
        self.step[test].status = settings.MONITOR.STATUS.ERROR
        self.step[test].error = self._exc_info_to_string(error, test)

    @monitors_step_required
    def addFailure(self, test, error):
        super().addFailure(test, error)
        self.step[test].status = settings.MONITOR.STATUS.FAILURE
        self.step[test].error = self._exc_info_to_string(error, test)
        self.step[test].reason = str(error[1])

    @monitors_step_required
    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self.step[test].status = settings.MONITOR.STATUS.SKIPPED
        self.step[test].reason = reason

    @monitors_step_required
    def addExpectedFailure(self, test, error):
        super().addExpectedFailure(test, error)
        self.step[test].status = settings.MONITOR.STATUS.EXPECTED_FAILURE
        self.step[test].error = self._exc_info_to_string(error, test)

    @monitors_step_required
    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        self.step[test].status = settings.MONITOR.STATUS.UNEXPECTED_SUCCESS

    @actions_step_required
    def start_action(self, action):
        self.step.add_item(action)

    @actions_step_required
    def add_action_success(self, action):
        self.step[action].status = settings.ACTION.STATUS.SUCCESS
        pass

    @actions_step_required
    def add_action_skip(self, action, reason):
        self.step[action].status = settings.ACTION.STATUS.SKIPPED
        self.step[action].reason = reason

    @actions_step_required
    def add_action_error(self, action, error):
        self.step[action].status = settings.ACTION.STATUS.ERROR
        self.step[action].error = error

    @actions_step_required
    def skip_all_step_actions(self, actions, reason):
        for action in actions:
            result = self.step.add_item(action)
            result.status = settings.ACTION.STATUS.SKIPPED
            result.reason = reason

    @property
    def _step_monitors(self):
        return self._steps[settings.STEPS.MONITORS]

    @property
    def _step_monitors_finished(self):
        return self._steps[settings.STEPS.MONITORS_FINISHED]

    @property
    def _step_monitors_passed(self):
        return self._steps[settings.STEPS.MONITORS_PASSED]

    @property
    def _step_monitors_failed(self):
        return self._steps[settings.STEPS.MONITORS_FAILED]

    def _get_step_class(self, step):
        return MonitorStep if step in settings.STEPS.MONITOR_RELATED else ActionsStep



import json
from io import BytesIO
from collections import defaultdict

from scrapy.exceptions import DropItem, NotConfigured
from scrapy.utils.misc import load_object
from scrapy.exporters import JsonLinesItemExporter
from scrapy import Field, Item
from scrapy.utils.python import to_unicode

from spidermon.contrib.validation import SchematicsValidator, JSONSchemaValidator
from spidermon.contrib.validation.jsonschema.tools import get_schema_from
from schematics.models import Model

from .stats import ValidationStatsManager


DEFAULT_ERRORS_FIELD = "_validation"
DEFAULT_ADD_ERRORS_TO_ITEM = False
DEFAULT_DROP_ITEMS_WITH_ERRORS = False


class PassThroughPipeline:
    def process_item(self, item, *args):
        return item


class ItemValidationPipeline:
    def __init__(
        self,
        validators,
        stats,
        drop_items_with_errors=DEFAULT_DROP_ITEMS_WITH_ERRORS,
        add_errors_to_items=DEFAULT_ADD_ERRORS_TO_ITEM,
        errors_field=None,
    ):
        self.drop_items_with_errors = drop_items_with_errors
        self.add_errors_to_items = add_errors_to_items or DEFAULT_ADD_ERRORS_TO_ITEM
        self.errors_field = errors_field or DEFAULT_ERRORS_FIELD
        self.validators = validators
        self.stats = ValidationStatsManager(stats)
        for _type, vals in validators.items():
            [self.stats.add_validator(_type, val.name) for val in vals]

    @classmethod
    def from_crawler(cls, crawler):
        spidermon_enabled = crawler.settings.getbool("SPIDERMON_ENABLED")
        if not spidermon_enabled:
            return PassThroughPipeline()

        validators = defaultdict(list)
        allowed_types = (list, tuple, dict)

        def set_validators(loader, schema):
            if type(schema) in (list, tuple):
                schema = {Item: schema}
            for obj, paths in schema.items():
                key = obj.__name__
                paths = paths if type(paths) in (list, tuple) else [paths]
                objects = [loader(v) for v in paths]
                validators[key].extend(objects)

        for loader, name in [
            (cls._load_jsonschema_validator, "SPIDERMON_VALIDATION_SCHEMAS"),
            (cls._load_schematics_validator, "SPIDERMON_VALIDATION_MODELS"),
        ]:
            res = crawler.settings.get(name)
            if not res:
                continue
            if type(res) not in allowed_types:
                raise NotConfigured(
                    "Invalid <{}> type for <{}> settings, dict or list/tuple"
                    "is required".format(type(res), name)
                )
            set_validators(loader, res)

        if not validators:
            raise NotConfigured("No validators were found")

        return cls(
            validators=validators,
            stats=crawler.stats,
            drop_items_with_errors=crawler.settings.getbool(
                "SPIDERMON_VALIDATION_DROP_ITEMS_WITH_ERRORS"
            ),
            add_errors_to_items=crawler.settings.getbool(
                "SPIDERMON_VALIDATION_ADD_ERRORS_TO_ITEMS"
            ),
            errors_field=crawler.settings.get("SPIDERMON_VALIDATION_ERRORS_FIELD"),
        )

    @classmethod
    def _load_jsonschema_validator(cls, schema):
        if isinstance(schema, str):
            schema = get_schema_from(schema)
        if not isinstance(schema, dict):
            raise NotConfigured(
                "Invalid schema, jsonschemas must be defined as:\n"
                "- a python dict.\n"
                "- an object path to a python dict.\n"
                "- an object path to a JSON string.\n"
                "- a path to a JSON file."
            )
        return JSONSchemaValidator(schema)

    @classmethod
    def _load_schematics_validator(cls, model_path):
        model_class = load_object(model_path)
        if not issubclass(model_class, Model):
            raise NotConfigured(
                "Invalid model, models must subclass schematics.models.Model"
            )
        return SchematicsValidator(model_class)

    def process_item(self, item, _):
        validators = self.find_validators(item)
        if not validators:
            # No validators match this specific item type
            return item

        data = self._convert_item_to_dict(item)
        self.stats.add_item()
        self.stats.add_fields(len(list(data.keys())))
        for validator in validators:
            ok, errors = validator.validate(data)
            if not ok:
                self._add_error_stats(errors)
                if self.add_errors_to_items:
                    self._add_errors_to_item(item, errors)
                if self.drop_items_with_errors:
                    self._drop_item(item, errors)
        return item

    def find_validators(self, item):
        find = lambda x: self.validators.get(x.__name__, [])
        return find(item.__class__) or find(Item)

    def _convert_item_to_dict(self, item):
        serialized_json = BytesIO()
        exporter = JsonLinesItemExporter(serialized_json)
        exporter.export_item(item)
        data = json.loads(to_unicode(serialized_json.getvalue(), exporter.encoding))
        serialized_json.close()
        return data

    def _add_errors_to_item(self, item, errors):
        try:
            if self.errors_field not in item.__class__.fields:
                item.__class__.fields[self.errors_field] = Field()
            if self.errors_field not in item._values:
                item[self.errors_field] = defaultdict(list)
        except AttributeError:
            # The item is just a dict object instead of a Scrapy.Item object
            if self.errors_field not in item:
                item[self.errors_field] = defaultdict(list)
        for field_name, messages in errors.items():
            item[self.errors_field][field_name] += messages

    def _drop_item(self, item, errors):
        """
        This method drops the item after detecting validation errors. Note
        that you could override it to add more details about the item that
        is being dropped or to drop the item only when some specific errors
        are detected.
        """
        self.stats.add_dropped_item()
        raise DropItem("Validation failed!")

    def _add_error_stats(self, errors):
        """
        This method adds validation error stats that can be later used to
        detect alert conditions in the monitors.
        """
        for field_name, messages in errors.items():
            for message in messages:
                self.stats.add_field_error(field_name, message)
        self.stats.add_item_with_errors()



from spidermon import MonitorSuite


from .cases import *


class EmptySuite(MonitorSuite):
    pass


class Suite01(MonitorSuite):
    monitors = [Monitor01]


class Suite02(MonitorSuite):
    monitors = [Suite01, Monitor02]


class Suite03(MonitorSuite):
    monitors = [Suite01, Suite02]


class Suite04(MonitorSuite):
    monitors = [Suite01, Suite02, Monitor01, Monitor02]



from unittest import TestSuite
import collections

from spidermon.exceptions import InvalidMonitorIterable, NotAllowedMethod
from spidermon import settings

from .monitors import Monitor
from .options import MonitorOptionsMetaclass
from .factories import MonitorFactory, ActionFactory


class MonitorSuite(TestSuite, metaclass=MonitorOptionsMetaclass):
    monitors = []
    monitors_finished_actions = []
    monitors_passed_actions = []
    monitors_failed_actions = []

    def __init__(
        self,
        name=None,
        monitors=None,
        monitors_finished_actions=None,
        monitors_passed_actions=None,
        monitors_failed_actions=None,
        order=None,
        crawler=None,
    ):
        self._tests = []
        self._removed_tests = 0
        self._name = name
        self._parent = None
        self._order = order
        self._crawler = crawler

        self.add_monitors(self.monitors)
        self.add_monitors(monitors or [])

        declarative_monitors_finished_actions = self.monitors_finished_actions
        self.monitors_finished_actions = []
        self.add_monitors_finished_actions(declarative_monitors_finished_actions)
        self.add_monitors_finished_actions(monitors_finished_actions or [])

        declarative_monitors_passed_actions = self.monitors_passed_actions
        self.monitors_passed_actions = []
        self.add_monitors_passed_actions(declarative_monitors_passed_actions)
        self.add_monitors_passed_actions(monitors_passed_actions or [])

        declarative_monitors_failed_actions = self.monitors_failed_actions
        self.monitors_failed_actions = []
        self.add_monitors_failed_actions(declarative_monitors_failed_actions)
        self.add_monitors_failed_actions(monitors_failed_actions or [])

    @property
    def name(self):
        return self._name or self.options.name or self.__class__.__name__

    @property
    def level(self):
        return self.options.level or self.parent_level

    @property
    def parent_level(self):
        if self.parent:
            return self.parent.level
        return settings.MONITOR.LEVELS.DEFAULT

    @property
    def full_name(self):
        parts = []
        if self.parent and self.parent.full_name:
            parts.append(self.parent.full_name)
        if self.have_custom_name:
            parts.append(self.name)
        return "/".join(parts)

    @property
    def have_custom_name(self):
        return self._name or self.options.name

    @property
    def description(self):
        return (
            self.options.description
            or self.__class__.__doc__
            or settings.MONITOR.DEFAULT_DESCRIPTION
        )

    @property
    def parent(self):
        return self._parent

    @property
    def order(self):
        return self._order if self._order is not None else None or self.options.order

    @property
    def number_of_monitors(self):
        return sum(
            [
                1 if isinstance(monitor, Monitor) else monitor.number_of_monitors
                for monitor in self
            ]
        )

    @property
    def all_monitors(self):
        monitors = []
        for monitor in self:
            if isinstance(monitor, Monitor):
                monitors += [monitor]
            else:
                monitors += monitor.all_monitors
        return monitors

    def set_parent(self, parent):
        self._parent = parent

    def init_data(self, data):
        for test in self:
            test.init_data(data)

    def add_monitors(self, monitors):
        if not isinstance(monitors, collections.abc.Iterable):
            raise InvalidMonitorIterable("Monitors definition is not iterable")
        for m in monitors:
            self.add_monitor(m)

    def add_monitor(self, monitor, name=None):
        monitor = MonitorFactory.load_monitor(monitor, name)
        monitor.set_parent(self)
        super().addTest(monitor)
        self._reorder_tests()

    def add_monitors_finished_actions(self, actions):
        for action in actions:
            self.add_monitors_finished_action(action)

    def add_monitors_finished_action(self, action):
        self._add_action(action, self.monitors_finished_actions)

    def add_monitors_passed_actions(self, actions):
        for action in actions:
            self.add_monitors_passed_action(action)

    def add_monitors_passed_action(self, action):
        self._add_action(action, self.monitors_passed_actions)

    def add_monitors_failed_actions(self, actions):
        for action in actions:
            self.add_monitors_failed_action(action)

    def add_monitors_failed_action(self, action):
        self._add_action(action, self.monitors_failed_actions)

    def _add_action(self, action, target_actions_list):
        action = ActionFactory.load_action(action, crawler=self._crawler)
        target_actions_list.append(action)

    def debug_tree(self, level=0):
        s = level * "\t" + repr(self) + "\n"
        for test in self:
            s += test.debug_tree(level=level + 1)
        return s

    def debug_monitors(
        self,
        show_monitor=True,
        show_method=True,
        show_level=True,
        show_order=False,
        show_description=True,
    ):
        def debug_attribute(condition, name, value):
            return "{:>12}: {}\n".format(name, str(value)) if condition else ""

        s = "-" * 80 + "\n"
        for t in self.all_monitors:
            s += debug_attribute(show_monitor, "MONITOR", t.monitor_full_name)
            s += debug_attribute(show_method, "METHOD", t.method_name)
            s += debug_attribute(show_level, "LEVEL", t.level)
            s += debug_attribute(show_order, "ORDER", t.order)
            s += debug_attribute(
                show_description, "DESCRIPTION", t.method_description or "..."
            )
            s += "-" * 80 + "\n"
        return s

    def _reorder_tests(self):
        self._tests = sorted(self._tests, key=lambda x: x.order, reverse=False)

    def __repr__(self):
        return "<SUITE:%s[%d,%s] at %s>" % (
            self.name,
            len(self._tests),
            self.number_of_monitors,
            hex(id(self)),
        )

    def __str__(self):
        return self.name

    def __not_allowed_method(self, *args, **kwargs):
        raise NotAllowedMethod

    addTest = __not_allowed_method
    addTests = __not_allowed_method

    def on_monitors_finished(self, result):
        pass

    def on_monitors_passed(self, result):
        pass

    def on_monitors_failed(self, result):
        pass



import datetime
import inspect
import os
import pprint as pretty_print

import jinja2
from jinja2 import Environment, FileSystemLoader

DEFAULT_TEMPLATE_FOLDERS = ["templates"]


def get_log_errors(logs):
    return [e for e in logs.list() if e["level"] >= 40]


def make_list(obj):
    return list(obj)


def pprint(obj):
    return pretty_print.pformat(obj)


def format_time(time):
    if not isinstance(time, datetime.timedelta):
        time = datetime.timedelta(seconds=int(time / 1000.0))
    return ":".join(str(time).split(":")[:2]) + "h"


FILTERS = {
    "pprint": pprint,
    "list": make_list,
    "get_log_errors": get_log_errors,
    "format_time": format_time,
}
GLOBALS = {"datetime": datetime, "str": str}


def get_environment(paths):
    loader = FileSystemLoader(paths)
    environment = Environment(loader=loader, lstrip_blocks=True, trim_blocks=True)
    for filter_name, filter in FILTERS.items():
        environment.filters[filter_name] = filter

    for global_name, global_value in GLOBALS.items():
        environment.globals[global_name] = global_value

    return environment


class TemplateLoader:
    def __init__(self):
        self.paths = []
        self.reload_env()

    def add_path(self, path):
        if path not in self.paths and os.path.isdir(path):
            self.paths.append(path)
            self.reload_env()

    def auto_discover(self, path=None, folder=None):
        caller_folder = os.path.dirname(inspect.stack()[1][1])
        if path:
            caller_folder = os.path.join(caller_folder, path)
        if folder:
            self.add_path(os.path.join(caller_folder, folder))
        else:
            self.discover_folder(caller_folder)

    def discover_folder(self, candidate_folder):
        for folder in [
            os.path.join(candidate_folder, dir) for dir in DEFAULT_TEMPLATE_FOLDERS
        ]:
            self.add_path(folder)

    def reload_env(self):
        self.env = get_environment(self.paths)

    def get_template(self, name):
        if os.path.isabs(name):  # If provided an absolute path to a template
            environment = get_environment(os.path.dirname(name))
            template = environment.get_template(os.path.basename(name))
        else:
            template = self.env.get_template(name)
        return template


template_loader = TemplateLoader()



STATS_A = {"item_scraped_count": 150}

STATS_B = {"item_scraped_count": 0}

STATS_EMPTY = {}

STATS_TO_EVALUATE = {
    "downloader/exception_count": 16,
    "downloader/response_count": 63785,
    "finish_reason": "finished",
    "item_scraped_count": 29836,
    "has_errors": True,
    "has_redirections": True,
}



import re


class MessageTranslator:

    messages = {}

    def __init__(self):
        self.compiled_messages = {m: re.compile(m) for m in self.messages}

    def translate_messages(self, messages):
        return [self.translate_message(m) for m in messages]

    def translate_message(self, message):
        for target_message, pattern in self.compiled_messages.items():
            pattern_found = pattern.search(message)
            if pattern_found:
                groups = pattern_found.groupdict()
                return self.messages[target_message].format(**groups)
        return message



from jsonschema._format import FormatChecker, _checks_drafts
from jsonschema.compat import str_types

from spidermon.contrib.validation.utils import is_valid_url, is_valid_email


@_checks_drafts("url")
def is_url(instance):
    if not isinstance(instance, str_types):
        return True
    return is_valid_url(instance)


@_checks_drafts("email")
def is_email(instance):
    if not isinstance(instance, str_types):
        return True
    return is_valid_email(instance)


format_checker = FormatChecker()



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import logging

logger = logging.getLogger(__name__)

from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured

from premailer import transform


class SendEmail(ActionWithTemplates):
    sender = None
    subject = None
    subject_template = None
    to = None
    cc = None
    bcc = None
    reply_to = None
    body_text = None
    body_text_template = None
    body_html = None
    body_html_template = "reports/email/monitors/result.jinja"
    fake = False

    def __init__(
        self,
        sender,
        to,
        cc=None,
        bcc=None,
        reply_to=None,
        subject=None,
        subject_template=None,
        body_text=None,
        body_text_template=None,
        body_html=None,
        body_html_template=None,
        fake=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sender = sender or self.sender
        self.subject = subject or self.subject
        self.subject_template = subject_template or self.subject_template
        self.to = to or self.to
        self.cc = cc or self.cc
        self.bcc = bcc or self.bcc
        self.reply_to = reply_to or self.reply_to
        self.body_text = body_text or self.body_text
        self.body_text_template = body_text_template or self.body_text_template
        self.body_html = body_html or self.body_html
        self.body_html_template = body_html_template or self.body_html_template
        self.fake = fake or self.fake
        if not self.fake and not self.to:
            raise NotConfigured(
                "You must provide at least one recipient for the message."
            )
        if not self.subject:
            raise NotConfigured("You must provide a subject for the message.")
        if not (
            self.body_text or body_text_template or body_html or self.body_html_template
        ):
            raise NotConfigured("You must provide a body for the message.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "sender": crawler.settings.get("SPIDERMON_EMAIL_SENDER"),
            "subject": crawler.settings.get("SPIDERMON_EMAIL_SUBJECT"),
            "subject_template": crawler.settings.get(
                "SPIDERMON_EMAIL_SUBJECT_TEMPLATE"
            ),
            "to": cls.getlist(crawler.settings, "SPIDERMON_EMAIL_TO"),
            "cc": cls.getlist(crawler.settings, "SPIDERMON_EMAIL_CC"),
            "bcc": cls.getlist(crawler.settings, "SPIDERMON_EMAIL_BCC"),
            "reply_to": crawler.settings.get("SPIDERMON_EMAIL_REPLY_TO"),
            "body_text": crawler.settings.get("SPIDERMON_BODY_TEXT"),
            "body_text_template": crawler.settings.get("SPIDERMON_BODY_TEXT_TEMPLATE"),
            "body_html": crawler.settings.get("SPIDERMON_BODY_HTML"),
            "body_html_template": crawler.settings.get("SPIDERMON_BODY_HTML_TEMPLATE"),
            "fake": crawler.settings.getbool("SPIDERMON_EMAIL_FAKE"),
            "context": crawler.settings.getdict("SPIDERMON_EMAIL_CONTEXT"),
        }

    @staticmethod
    def getlist(settings, entry):
        return [v.strip() for v in settings.getlist(entry)]

    def run_action(self):
        message = self.get_message()
        if self.fake:
            logger.info(message.as_string())
        else:
            self.send_message(message)

    def get_subject(self):
        if self.subject:
            return self.render_text_template(self.subject)
        elif self.subject_template:
            return self.render_template(self.subject_template)
        else:
            return ""

    def get_body_text(self):
        if self.body_text:
            return self.render_text_template(self.body_text)
        elif self.body_text_template:
            return self.render_template(self.body_text_template)
        else:
            return ""

    def get_body_html(self):
        html = ""
        if self.body_html:
            html = transform(self.render_text_template(self.body_html))
        elif self.body_html_template:
            html = transform(self.render_template(self.body_html_template))
        return html

    def get_message(self):
        subject = self.get_subject()
        body_text = self.get_body_text()
        body_html = self.get_body_html()

        message = MIMEMultipart("alternative")
        message.set_charset("UTF-8")

        message["Subject"] = subject
        message["From"] = self.sender
        message["To"] = ", ".join(self.to)
        if self.cc:
            message["Cc"] = ", ".join(self.cc)
        if self.bcc:
            message["Bcc"] = ", ".join(self.bcc)
        if self.reply_to:
            message["reply-to"] = self.reply_to

        message.attach(MIMEText(body_text, "plain"))
        if body_html:
            message.attach(MIMEText(body_html, "html"))

        return message

    def send_message(self, message):
        raise NotImplementedError



from .fixtures.descriptions import *

MONITOR_DESCRIPTIONS = [
    # --------------------------------------------------------------
    # monitor class                 description
    # --------------------------------------------------------------
    (NotDescriptedMonitor, ""),
    (DocstringDescriptedMonitor, "docstring monitor description"),
    (DescoratedDescriptedMonitor, "decorator monitor description"),
    (DescoratedDescriptedMonitor2, "decorator monitor description"),
]

SUITE_DESCRIPTIONS = [
    # --------------------------------------------------------------
    # suite class                   description
    # --------------------------------------------------------------
    (NotDescriptedSuite, ""),
    (DocstringDescriptedSuite, "docstring suite description"),
    (DescoratedDescriptedSuite, "decorator suite description"),
    (DescoratedDescriptedSuite2, "decorator suite description"),
]

METHOD_DESCRIPTIONS = [
    # ---------------------------------------------------------------------------------------------
    # monitor class                 method name                     description
    # ---------------------------------------------------------------------------------------------
    (DescriptedMethodsMonitor, "test_not_descripted", ""),
    (
        DescriptedMethodsMonitor,
        "test_docstring_descripted",
        "docstring method description",
    ),
    (
        DescriptedMethodsMonitor,
        "test_decorator_descripted",
        "decorator method description",
    ),
    (
        DescriptedMethodsMonitor,
        "test_decorator_descripted2",
        "decorator method description",
    ),
]


def test_monitor_descriptions():
    for monitor_class, description in MONITOR_DESCRIPTIONS:
        assert monitor_class().monitor_description == description


def test_suite_descriptions():
    for suite_class, description in SUITE_DESCRIPTIONS:
        assert suite_class().description == description


def test_method_descriptions():
    for monitor_class, method_name, description in METHOD_DESCRIPTIONS:
        assert monitor_class(method_name).method_description == description



import inspect
from unittest import TestLoader
from functools import cmp_to_key as _cmp_to_key

from .core.monitors import Monitor
from .core.suites import MonitorSuite
from .exceptions import InvalidMonitor


class MonitorLoader(TestLoader):
    def load_suite_from_monitor(self, monitor_class, name=None):
        if not (inspect.isclass(monitor_class) and issubclass(monitor_class, Monitor)):
            raise InvalidMonitor("monitor must be a class subclassing Monitor")
        test_function_names = self.get_testcase_names(monitor_class)
        if not test_function_names and hasattr(monitor_class, "runTest"):
            test_function_names = ["runTest"]
        monitors = [
            monitor_class(fn_name, name=name) for fn_name in test_function_names
        ]
        loaded_suite = MonitorSuite(
            monitors=monitors, order=monitor_class.options.order
        )
        return loaded_suite

    def get_testcase_names(self, monitor_class):
        def is_test_method(
            attrname, class_name=monitor_class, prefix=self.testMethodPrefix
        ):
            return attrname.startswith(prefix) and hasattr(
                getattr(class_name, attrname), "__call__"
            )

        test_function_names = list(filter(is_test_method, dir(monitor_class)))
        if self.sortTestMethodsUsing:
            test_function_names.sort(key=_cmp_to_key(self.sortTestMethodsUsing))
        return test_function_names

    # TODO: hide methods?



from spidermon.core.options import ActionOptions
from spidermon.decorators import OptionsDecorator


name = OptionsDecorator.set_value(ActionOptions, "name")
description = OptionsDecorator.set_value(ActionOptions, "description")



__version__ = "1.16.2"

from .core.monitors import Monitor
from .core.suites import MonitorSuite
from .core.actions import Action, DummyAction
from .loaders import MonitorLoader
from .runners import MonitorRunner, TextMonitorRunner
from .results.monitor import MonitorResult
from .results.text import TextMonitorResult
from .decorators import monitors, actions
from .exceptions import SkipAction



from spidermon import settings

from .fixtures.levels import *

HIGH = settings.MONITOR.LEVEL.HIGH
NORMAL = settings.MONITOR.LEVEL.NORMAL
LOW = settings.MONITOR.LEVEL.LOW


LEVEL_TESTS = [
    # ---------------------------------------------------------------------------------------
    # suite                     monitor/methods                                expected level
    # ---------------------------------------------------------------------------------------
    # Suite No Level
    (Suites.NoLevelSuite, Monitors.NoLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.NoLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NoLevelSuite, Monitors.NoLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.NoLevelMonitor.LowLevelMethod, LOW),
    (Suites.NoLevelSuite, Monitors.HighLevelMonitor.NoLevelMethod, HIGH),
    (Suites.NoLevelSuite, Monitors.HighLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NoLevelSuite, Monitors.HighLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.HighLevelMonitor.LowLevelMethod, LOW),
    (Suites.NoLevelSuite, Monitors.NormalLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.NormalLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NoLevelSuite, Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.NormalLevelMonitor.LowLevelMethod, LOW),
    (Suites.NoLevelSuite, Monitors.LowLevelMonitor.NoLevelMethod, LOW),
    (Suites.NoLevelSuite, Monitors.LowLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NoLevelSuite, Monitors.LowLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NoLevelSuite, Monitors.LowLevelMonitor.LowLevelMethod, LOW),
    # Suite High
    (Suites.HighLevelSuite, Monitors.NoLevelMonitor.NoLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.NoLevelMonitor.HighLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.NoLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.HighLevelSuite, Monitors.NoLevelMonitor.LowLevelMethod, LOW),
    (Suites.HighLevelSuite, Monitors.HighLevelMonitor.NoLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.HighLevelMonitor.HighLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.HighLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.HighLevelSuite, Monitors.HighLevelMonitor.LowLevelMethod, LOW),
    (Suites.HighLevelSuite, Monitors.NormalLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.HighLevelSuite, Monitors.NormalLevelMonitor.HighLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.HighLevelSuite, Monitors.NormalLevelMonitor.LowLevelMethod, LOW),
    (Suites.HighLevelSuite, Monitors.LowLevelMonitor.NoLevelMethod, LOW),
    (Suites.HighLevelSuite, Monitors.LowLevelMonitor.HighLevelMethod, HIGH),
    (Suites.HighLevelSuite, Monitors.LowLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.HighLevelSuite, Monitors.LowLevelMonitor.LowLevelMethod, LOW),
    # Suite Normal
    (Suites.NormalLevelSuite, Monitors.NoLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.NoLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NormalLevelSuite, Monitors.NoLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.NoLevelMonitor.LowLevelMethod, LOW),
    (Suites.NormalLevelSuite, Monitors.HighLevelMonitor.NoLevelMethod, HIGH),
    (Suites.NormalLevelSuite, Monitors.HighLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NormalLevelSuite, Monitors.HighLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.HighLevelMonitor.LowLevelMethod, LOW),
    (Suites.NormalLevelSuite, Monitors.NormalLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.NormalLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NormalLevelSuite, Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.NormalLevelMonitor.LowLevelMethod, LOW),
    (Suites.NormalLevelSuite, Monitors.LowLevelMonitor.NoLevelMethod, LOW),
    (Suites.NormalLevelSuite, Monitors.LowLevelMonitor.HighLevelMethod, HIGH),
    (Suites.NormalLevelSuite, Monitors.LowLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NormalLevelSuite, Monitors.LowLevelMonitor.LowLevelMethod, LOW),
    # Suite Low
    (Suites.LowLevelSuite, Monitors.NoLevelMonitor.NoLevelMethod, LOW),
    (Suites.LowLevelSuite, Monitors.NoLevelMonitor.HighLevelMethod, HIGH),
    (Suites.LowLevelSuite, Monitors.NoLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.LowLevelSuite, Monitors.NoLevelMonitor.LowLevelMethod, LOW),
    (Suites.LowLevelSuite, Monitors.HighLevelMonitor.NoLevelMethod, HIGH),
    (Suites.LowLevelSuite, Monitors.HighLevelMonitor.HighLevelMethod, HIGH),
    (Suites.LowLevelSuite, Monitors.HighLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.LowLevelSuite, Monitors.HighLevelMonitor.LowLevelMethod, LOW),
    (Suites.LowLevelSuite, Monitors.NormalLevelMonitor.NoLevelMethod, NORMAL),
    (Suites.LowLevelSuite, Monitors.NormalLevelMonitor.HighLevelMethod, HIGH),
    (Suites.LowLevelSuite, Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.LowLevelSuite, Monitors.NormalLevelMonitor.LowLevelMethod, LOW),
    (Suites.LowLevelSuite, Monitors.LowLevelMonitor.NoLevelMethod, LOW),
    (Suites.LowLevelSuite, Monitors.LowLevelMonitor.HighLevelMethod, HIGH),
    (Suites.LowLevelSuite, Monitors.LowLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.LowLevelSuite, Monitors.LowLevelMonitor.LowLevelMethod, LOW),
]


def test_levels():
    for suite, monitor, expected_level in LEVEL_TESTS:
        suite = suite()
        suite.add_monitor(monitor)
        assert suite.all_monitors[0].level == expected_level



import pytest

from spidermon.data import Data
from spidermon.exceptions import InvalidDataOperation


@pytest.fixture
def data():
    return Data(item_scraped_count=150)


def test_attribute_access(data):
    assert data["item_scraped_count"] == 150


def test_dictionary_access(data):
    assert data.item_scraped_count == 150


def test_attribute_set(data):
    with pytest.raises(InvalidDataOperation):
        data.item_scraped_count = "some value"


def test_dictionary_set(data):
    with pytest.raises(InvalidDataOperation):
        data["item_scraped_count"] = "some value"


def test_delete(data):
    with pytest.raises(InvalidDataOperation):
        del data["item_scraped_count"]


def test_pop(data):
    with pytest.raises(InvalidDataOperation):
        data.pop("item_scraped_count", None)


def test_clear(data):
    with pytest.raises(InvalidDataOperation):
        data.clear()


def test_update(data):
    with pytest.raises(InvalidDataOperation):
        data.update({"item_scraped_count": 0})


def test_setdefault(data):
    with pytest.raises(InvalidDataOperation):
        data.setdefault("another_value", 0)



from spidermon.core.options import MonitorOptions
from spidermon import settings
from spidermon.decorators import OptionsDecorator, DecoratorWithAttributes


class LevelDecorator(DecoratorWithAttributes):
    name = "level"
    attributes = {
        "high": OptionsDecorator.set_fixed_value(
            MonitorOptions, name, settings.MONITOR.LEVEL.HIGH
        ),
        "normal": OptionsDecorator.set_fixed_value(
            MonitorOptions, name, settings.MONITOR.LEVEL.NORMAL
        ),
        "low": OptionsDecorator.set_fixed_value(
            MonitorOptions, name, settings.MONITOR.LEVEL.LOW
        ),
    }


name = OptionsDecorator.set_value(MonitorOptions, "name")
description = OptionsDecorator.set_value(MonitorOptions, "description")
order = OptionsDecorator.set_value(MonitorOptions, "order")
level = LevelDecorator()



from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured


class CreateReport(ActionWithTemplates):
    template = None

    def __init__(self, template=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template = template or self.template
        self.report = ""
        if not self.template:
            raise NotConfigured("You must define one template file.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        kwargs.update(
            {
                "template": crawler.settings.get("SPIDERMON_REPORT_TEMPLATE"),
                "context": crawler.settings.getdict("SPIDERMON_REPORT_CONTEXT"),
            }
        )
        return kwargs

    def run_action(self):
        self.before_render_report()
        self.render_report()
        self.after_render_report()

    def before_render_report(self):
        pass

    def render_report(self):
        self.report = self.render_template(self.template)

    def after_render_report(self):
        pass



from spidermon import Monitor, MonitorSuite, monitors


# ----------------------------------
# Base Monitors
# ----------------------------------
class NoLevelMethodMonitor(Monitor):
    def test(self):
        pass


class HighLevelMethodMonitor(Monitor):
    @monitors.level.high
    def test(self):
        pass


class NormalLevelMethodMonitor(Monitor):
    @monitors.level.normal
    def test(self):
        pass


class LowLevelMethodMonitor(Monitor):
    @monitors.level.low
    def test(self):
        pass


# ----------------------------------
# Monitors
# ----------------------------------
class Monitors:
    class NoLevelMonitor:
        class NoLevelMethod(NoLevelMethodMonitor):
            pass

        class HighLevelMethod(HighLevelMethodMonitor):
            pass

        class NormalLevelMethod(NormalLevelMethodMonitor):
            pass

        class LowLevelMethod(LowLevelMethodMonitor):
            pass

    class HighLevelMonitor:
        @monitors.level.high
        class NoLevelMethod(NoLevelMethodMonitor):
            pass

        @monitors.level.high
        class HighLevelMethod(HighLevelMethodMonitor):
            pass

        @monitors.level.high
        class NormalLevelMethod(NormalLevelMethodMonitor):
            pass

        @monitors.level.high
        class LowLevelMethod(LowLevelMethodMonitor):
            pass

    class NormalLevelMonitor:
        @monitors.level.normal
        class NoLevelMethod(NoLevelMethodMonitor):
            pass

        @monitors.level.normal
        class HighLevelMethod(HighLevelMethodMonitor):
            pass

        @monitors.level.normal
        class NormalLevelMethod(NormalLevelMethodMonitor):
            pass

        @monitors.level.normal
        class LowLevelMethod(LowLevelMethodMonitor):
            pass

    class LowLevelMonitor:
        @monitors.level.low
        class NoLevelMethod(NoLevelMethodMonitor):
            pass

        @monitors.level.low
        class HighLevelMethod(HighLevelMethodMonitor):
            pass

        @monitors.level.low
        class NormalLevelMethod(NormalLevelMethodMonitor):
            pass

        @monitors.level.low
        class LowLevelMethod(LowLevelMethodMonitor):
            pass


# ----------------------------------
# Suites
# ----------------------------------
class Suites:
    class NoLevelSuite(MonitorSuite):
        pass

    @monitors.level.high
    class HighLevelSuite(MonitorSuite):
        pass

    @monitors.level.normal
    class NormalLevelSuite(MonitorSuite):
        pass

    @monitors.level.low
    class LowLevelSuite(MonitorSuite):
        pass



from collections import OrderedDict
import time

from spidermon import settings

from .items import ItemResult, MonitorResult, ActionResult


class Step:
    item_result_class = ItemResult
    successful_statuses = []
    error_statuses = []

    def __init__(self, name):
        self.name = name
        self._results = OrderedDict()
        self.start_time = 0
        self.finish_time = 0

    @property
    def time_taken(self):
        return self.finish_time - self.start_time

    @property
    def number_of_items(self):
        return len(self._results)

    def add_item(self, item):
        result = self.item_result_class(item)
        self._results[item] = result
        return result

    def start(self):
        self.start_time = time.time()

    def finish(self):
        self.finish_time = time.time()

    def items_for_status(self, status):
        return [
            result for item, result in self._results.items() if result.status == status
        ]

    def items_for_statuses(self, statuses):
        items = []
        for status in statuses:
            items += self.items_for_status(status)
        return items

    @property
    def all_items(self):
        return list(self._results.values())

    def get_infos(self):
        raise NotImplementedError

    def __getitem__(self, key):
        return self._results[key]

    @property
    def successful_results(self):
        results = []
        for successful_status in self.successful_statuses:
            results += self.items_for_status(successful_status)
        return results

    @property
    def error_results(self):
        results = []
        for error_status in self.error_statuses:
            results += self.items_for_status(error_status)
        return results

    @property
    def successful(self):
        return not self.has_errors

    @property
    def has_errors(self):
        return len(self.error_results) > 0


class MonitorStep(Step):
    item_result_class = MonitorResult
    successful_statuses = settings.MONITOR.STATUSES.SUCCESSFUL
    error_statuses = settings.MONITOR.STATUSES.ERROR

    def get_infos(self):
        return {
            "failures": len(self.items_for_status(settings.MONITOR.STATUS.FAILURE)),
            "errors": len(self.items_for_status(settings.MONITOR.STATUS.ERROR)),
            "skipped": len(self.items_for_status(settings.MONITOR.STATUS.SKIPPED)),
            "expected failures": len(
                self.items_for_status(settings.MONITOR.STATUS.EXPECTED_FAILURE)
            ),
            "unexpected successes": len(
                self.items_for_status(settings.MONITOR.STATUS.UNEXPECTED_SUCCESS)
            ),
        }


class ActionsStep(Step):
    item_result_class = ActionResult
    successful_statuses = settings.ACTION.STATUSES.SUCCESSFUL
    error_statuses = settings.ACTION.STATUSES.ERROR

    def get_infos(self):
        return {
            "errors": len(self.items_for_status(settings.ACTION.STATUS.ERROR)),
            "skipped": len(self.items_for_status(settings.ACTION.STATUS.SKIPPED)),
        }



import os
import pickle
from collections import deque

from scrapy.statscollectors import StatsCollector
from scrapy.utils.project import data_path


class LocalStorageStatsHistoryCollector(StatsCollector):
    def _stats_location(self, spider):
        statsdir = data_path("stats", createdir=True)
        return os.path.join(statsdir, f"{spider.name}_stats_history")

    def open_spider(self, spider):
        stats_location = self._stats_location(spider)

        max_stored_stats = spider.crawler.settings.getint(
            "SPIDERMON_MAX_STORED_STATS", default=100
        )

        if os.path.isfile(stats_location):
            with open(stats_location, "rb") as stats_file:
                _stats_history = pickle.load(stats_file)
        else:
            _stats_history = deque([], maxlen=max_stored_stats)

        if _stats_history.maxlen != max_stored_stats:
            _stats_history = deque(_stats_history, maxlen=max_stored_stats)

        spider.stats_history = _stats_history

    def _persist_stats(self, stats, spider):
        stats_location = self._stats_location(spider)

        spider.stats_history.appendleft(self._stats)
        with open(stats_location, "wb") as stats_file:
            pickle.dump(spider.stats_history, stats_file)



class SpidermonException(Exception):
    pass


class InvalidDataOperation(SpidermonException):
    pass


class NotAllowedMethod(SpidermonException):
    pass


class InvalidMonitor(SpidermonException):
    pass


class InvalidMonitorIterable(InvalidMonitor):
    pass


class InvalidMonitorClass(InvalidMonitor):
    pass


class InvalidMonitorTuple(InvalidMonitor):
    pass


class InvalidExpression(SpidermonException):
    pass


class InvalidResult(SpidermonException):
    pass


class NotConfigured(SpidermonException):
    pass


class SkipAction(SpidermonException):
    pass



# ----------------------------------
# Status
# ----------------------------------
UNDEFINED_STATUS = "?"

# ----------------------------------
# Monitors
# ----------------------------------
_MONITOR_STATUS_SUCCESS = "OK"
_MONITOR_STATUS_FAILURE = "FAIL"
_MONITOR_STATUS_SKIPPED = "SKIPPED"
_MONITOR_STATUS_ERROR = "ERROR"
_MONITOR_STATUS_EXPECTED_FAILURE = "EXPECTED FAILURE"
_MONITOR_STATUS_UNEXPECTED_SUCCESS = "UNEXPECTED SUCCESS"

_MONITOR_LEVEL_HIGH = "HIGH"
_MONITOR_LEVEL_NORMAL = "NORMAL"
_MONITOR_LEVEL_LOW = "LOW"


class MONITOR:
    class STATUS:
        SUCCESS = _MONITOR_STATUS_SUCCESS
        FAILURE = _MONITOR_STATUS_FAILURE
        SKIPPED = _MONITOR_STATUS_SKIPPED
        ERROR = _MONITOR_STATUS_ERROR
        EXPECTED_FAILURE = _MONITOR_STATUS_EXPECTED_FAILURE
        UNEXPECTED_SUCCESS = _MONITOR_STATUS_UNEXPECTED_SUCCESS

    class STATUSES:
        ALL = (
            _MONITOR_STATUS_SUCCESS,
            _MONITOR_STATUS_FAILURE,
            _MONITOR_STATUS_SKIPPED,
            _MONITOR_STATUS_ERROR,
            _MONITOR_STATUS_EXPECTED_FAILURE,
            _MONITOR_STATUS_UNEXPECTED_SUCCESS,
        )
        SUCCESSFUL = (
            _MONITOR_STATUS_SUCCESS,
            _MONITOR_STATUS_SKIPPED,
            _MONITOR_STATUS_EXPECTED_FAILURE,
            _MONITOR_STATUS_UNEXPECTED_SUCCESS,
        )
        ERROR = (_MONITOR_STATUS_FAILURE, _MONITOR_STATUS_ERROR)
        DEFAULT = UNDEFINED_STATUS

    class LEVEL:
        HIGH = _MONITOR_LEVEL_HIGH
        NORMAL = _MONITOR_LEVEL_NORMAL
        LOW = _MONITOR_LEVEL_LOW

    class LEVELS:
        ALL = (_MONITOR_LEVEL_HIGH, _MONITOR_LEVEL_NORMAL, _MONITOR_LEVEL_LOW)
        DEFAULT = _MONITOR_LEVEL_NORMAL

    DEFAULT_NAME = None
    DEFAULT_DESCRIPTION = ""
    DEFAULT_ORDER = 1


# ----------------------------------
# Actions
# ----------------------------------
_ACTION_STATUS_SUCCESS = "OK"
_ACTION_STATUS_ERROR = "ERROR"
_ACTION_STATUS_SKIPPED = "SKIPPED"


class ACTION:
    class STATUS:
        SUCCESS = _ACTION_STATUS_SUCCESS
        ERROR = _ACTION_STATUS_ERROR
        SKIPPED = _ACTION_STATUS_SKIPPED

    class STATUSES:
        ALL = (_ACTION_STATUS_SUCCESS, _ACTION_STATUS_ERROR, _ACTION_STATUS_SKIPPED)
        SUCCESSFUL = (_ACTION_STATUS_SUCCESS, _ACTION_STATUS_SKIPPED)
        ERROR = (_ACTION_STATUS_ERROR,)
        DEFAULT = UNDEFINED_STATUS

    DEFAULT_DESCRIPTION = ""


# ----------------------------------
# Steps
# ----------------------------------
_STEP_MONITORS = "MONITORS"
_STEP_MONITORS_FINISHED = "FINISHED ACTIONS"
_STEP_MONITORS_PASSED = "PASSED ACTIONS"
_STEP_MONITORS_FAILED = "FAILED ACTIONS"


class STEPS:
    MONITORS = _STEP_MONITORS
    MONITORS_FINISHED = _STEP_MONITORS_FINISHED
    MONITORS_PASSED = _STEP_MONITORS_PASSED
    MONITORS_FAILED = _STEP_MONITORS_FAILED
    ALL = (
        _STEP_MONITORS,
        _STEP_MONITORS_FINISHED,
        _STEP_MONITORS_PASSED,
        _STEP_MONITORS_FAILED,
    )
    MONITOR_RELATED = (_STEP_MONITORS,)
    ACTION_RELATED = (
        _STEP_MONITORS_FINISHED,
        _STEP_MONITORS_PASSED,
        _STEP_MONITORS_FAILED,
    )



import json
from schematics.models import Model
from schematics.types import URLType, StringType, BaseType


class TestValidator(Model):
    url = URLType(required=True)
    title = StringType()


class TreeValidator(Model):
    child = BaseType(required=True)


tree_schema = {
    "$schema": "http://json-schema.org/draft-04/schema",
    "required": ["child"],
    "type": "object",
    "properties": {"child": {"type": "object"}},
}

test_schema = {
    "$schema": "http://json-schema.org/draft-04/schema",
    "required": ["url"],
    "type": "object",
    "properties": {"url": {"type": "string"}, "title": {"type": "string"}},
}

test_schema_string = json.dumps(test_schema)



import schematics


def monkeypatch_urltype():
    """
    Replace schematics URL check regex with a better one (stolen from Django).

    This patch cannot be applied to Schematics 2.* because the URL validation
    is more complex.
    """
    from schematics.types import URLType
    from spidermon.contrib.validation.utils import URL_REGEX

    URLType.URL_REGEX = URL_REGEX


def monkeypatch_listtype():
    """
    Replace ListType list conversion method to avoid errors
    """
    from schematics.transforms import EMPTY_LIST
    from schematics.types.compound import ListType
    from schematics.exceptions import ConversionError

    def _force_list(self, value):
        if value is None or value == EMPTY_LIST:
            return []
        try:
            return list(value)
        except Exception as e:
            raise ConversionError("Invalid list")

    ListType._force_list = _force_list


# Apply monkeypatches
if schematics.__version__.startswith("1."):
    monkeypatch_urltype()
    monkeypatch_listtype()



import re

import schematics
from schematics.exceptions import ModelValidationError, ModelConversionError

from spidermon.contrib.validation.validator import Validator
from .translator import SchematicsMessageTranslator
from . import monkeypatches


class SchematicsValidator(Validator):
    default_translator = SchematicsMessageTranslator()
    name = "Schematics"

    def __init__(self, model, translator=None, use_default_translator=True):
        super().__init__(
            translator=translator, use_default_translator=use_default_translator
        )
        self._model = model
        self._fields_required = {}
        self._save_required_fields()
        self._data = {}

    def _validate(self, data, strict=False):
        self._set_data(data)
        model = self._get_model_instance(strict=strict)
        try:
            model.validate()
        except ModelValidationError as e:
            self._add_errors(e.messages)
        self._restore_required_fields()

    def _reset(self):
        super()._reset()
        self._data = {}

    def _set_data(self, data):
        self._data = dict(data)

    def _get_model_instance(self, strict):
        try:
            return self._model(raw_data=self._data, strict=strict)
        except ModelConversionError as e:
            self._add_errors(e.messages)
            for field_name in e.messages.keys():
                self._set_field_as_not_required(field_name)
                self._data.pop(field_name)
            return self._get_model_instance(strict=strict)

    def _save_required_fields(self):
        for field_name, field in self._model._fields.items():
            self._fields_required[field_name] = field.required

    def _restore_required_fields(self):
        for field_name, required in self._fields_required.items():
            self._model._fields[field_name].required = required

    def _set_field_as_not_required(self, field_name):
        if field_name in self._model._fields:
            self._model._fields[field_name].required = False

    def _add_errors(self, errors):
        if schematics.__version__.startswith("1."):
            for field_name, messages in errors.items():
                if isinstance(messages, dict):
                    transformed_errors = self._get_transformed_child_errors(
                        field_name, messages
                    )
                    self._add_errors(transformed_errors)
                else:
                    self._errors[field_name] += (
                        messages if isinstance(messages, list) else [messages]
                    )
        else:
            from schematics.datastructures import FrozenDict

            for field_name, messages in errors.items():
                if isinstance(messages, (dict, FrozenDict)):
                    transformed_errors = self._get_transformed_child_errors(
                        field_name, messages
                    )
                    self._add_errors(transformed_errors)
                else:
                    messages = self._clean_messages(messages)
                    self._errors[field_name] += messages

    def _get_transformed_child_errors(self, field_name, errors):
        return {f"{field_name}.{k}": v for k, v in errors.items()}

    def _clean_messages(self, messages):
        """
        This is necessary when using Schematics 2.*, because it encapsulates
        the validation error messages in a different way.
        """
        from schematics.exceptions import BaseError, ErrorMessage
        from schematics.datastructures import FrozenList

        if type(messages) not in (list, FrozenList):
            messages = [messages]

        clean_messages = []
        for message in messages:
            if isinstance(message, BaseError):
                message = message.messages

            if isinstance(message, ErrorMessage):
                clean_messages.append(message.summary)
            elif isinstance(message, FrozenList):
                for err in message:
                    # err is an ErrorMessage object
                    clean_messages.append(err.summary)
            else:
                clean_messages.append(message)

        return clean_messages



import sys

from spidermon import settings
from spidermon.utils.text import line_title

from .monitor import MonitorResult, monitors_step_required, actions_step_required


DOTS = {
    # Monitors
    settings.MONITOR.STATUS.SUCCESS: ".",
    settings.MONITOR.STATUS.ERROR: "E",
    settings.MONITOR.STATUS.FAILURE: "F",
    settings.MONITOR.STATUS.SKIPPED: "s",
    settings.MONITOR.STATUS.EXPECTED_FAILURE: "x",
    settings.MONITOR.STATUS.UNEXPECTED_SUCCESS: "u",
    # Actions
    settings.ACTION.STATUS.SUCCESS: ".",
    settings.ACTION.STATUS.ERROR: "E",
    settings.ACTION.STATUS.SKIPPED: "s",
}


class TextMonitorResult(MonitorResult):

    SEPARATOR_BOLD = "="
    SEPARATOR_LIGHT = "-"
    LINE_LENGTH = 70

    def __init__(self, stream=sys.stderr, verbosity=1):
        super().__init__()
        self.stream = stream
        self.show_all = verbosity > 1
        self.use_dots = verbosity == 1

    def next_step(self):
        super().next_step()
        self.write_title(self.step.name)

    def finish_step(self):
        super().finish_step()
        if self.use_dots:
            self.write_line()
        if not self.step.successful:
            self.write_errors()
        self.write_run_footer()
        self.write_step_summary()

    @monitors_step_required
    def startTest(self, test):
        super().startTest(test)
        self.write_run_start(test)

    @monitors_step_required
    def addSuccess(self, test):
        super().addSuccess(test)
        self.write_run_result(test)

    @monitors_step_required
    def addError(self, test, error):
        super().addError(test, error)
        self.write_run_result(test)

    @monitors_step_required
    def addFailure(self, test, error):
        super().addFailure(test, error)
        self.write_run_result(test)

    @monitors_step_required
    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self.write_run_result(test, reason)

    @monitors_step_required
    def addExpectedFailure(self, test, error):
        super().addExpectedFailure(test, error)
        self.write_run_result(test)

    @monitors_step_required
    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        self.write_run_result(test)

    @actions_step_required
    def start_action(self, action):
        super().start_action(action)
        self.write_run_start(action)

    @actions_step_required
    def add_action_success(self, action):
        super().add_action_success(action)
        self.write_run_result(action)

    @actions_step_required
    def add_action_skip(self, action, reason):
        super().add_action_skip(action, reason)
        self.write_run_result(action, reason)

    @actions_step_required
    def add_action_error(self, action, error):
        super().add_action_error(action, error)
        self.write_run_result(action)

    def write(self, text):
        self.stream.write(text)

    def write_flush(self):
        self.stream.flush()

    def write_line_light(self):
        self.write_line(self.SEPARATOR_LIGHT * self.LINE_LENGTH)

    def write_line_bold(self):
        self.write_line(self.SEPARATOR_BOLD * self.LINE_LENGTH)

    def write_title(self, title):
        self.write_line(line_title(title))

    def write_line(self, text=None):
        self.write("%s\n" % (text or ""))

    def write_run_status(self, text, extra=None):
        self.write_line("{}{}".format(text, " (%s)" % extra if extra else ""))

    def write_run_start(self, item):
        if self.show_all:
            self.write(item.name)
            self.write(" ... ")
            self.write_flush()

    def write_run_result(self, item, extra=None):
        if self.show_all:
            self.write_run_status(self.step[item].status, extra)
        elif self.use_dots:
            self.write(DOTS[self.step[item].status])
            self.write_flush()

    def write_run_footer(self):
        self.write_line_light()
        self.write_line(
            "{count:d} {item_name}{plural_suffix} in {time:.3f}s".format(
                count=self.step.number_of_items,
                item_name=self.step.item_result_class.name,
                plural_suffix="" if self.step.number_of_items == 1 else "s",
                time=self.step.time_taken,
            )
        )
        self.write_line()

    def write_errors(self):
        self.write_line()
        for status in self.step.error_statuses:
            for item in self.step.items_for_status(status):
                self.write_line_bold()
                self.write_line(f"{item.status}: {item.item.name}")
                self.write_line_light()
                self.write_line(item.error)
                self.write_line()

    def write_step_summary(self):
        self.write("OK" if self.step.successful else "FAILED")
        infos = self.step.get_infos()
        if infos and sum(infos.values()):
            self.write_line(
                " (%s)" % ", ".join([f"{k}={v}" for k, v in infos.items() if v])
            )
        else:
            self.write_line()
        self.write_line()



import copy
import json
from collections import OrderedDict
import warnings


def getdictorlist(crawler, name, default=None):
    value = crawler.settings.get(name, default)
    if value is None:
        return {}
    if isinstance(value, str):
        try:
            return json.loads(value, object_pairs_hook=OrderedDict)
        except ValueError:
            return value.split(",")
    return copy.deepcopy(value)


def get_aws_credentials(settings):
    aws_access_key_id = settings.get("SPIDERMON_AWS_ACCESS_KEY")
    aws_secret_access_key = settings.get("SPIDERMON_AWS_SECRET_KEY")

    if aws_access_key_id and aws_secret_access_key:
        warnings.warn(
            "SPIDERMON_AWS_ACCESS_KEY and SPIDERMON_AWS_SECRET_KEY are deprecated. "
            "Please update them to SPIDERMON_AWS_ACCESS_KEY_ID and SPIDERMON_AWS_SECRET_ACCESS_KEY. "  # noqa
            "Scrapy settings AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY are also valid.",
            DeprecationWarning,
        )

    if not aws_access_key_id and not aws_secret_access_key:
        aws_access_key_id = settings.get("SPIDERMON_AWS_ACCESS_KEY_ID")
        aws_secret_access_key = settings.get("SPIDERMON_AWS_SECRET_ACCESS_KEY")

    if not aws_access_key_id and not aws_secret_access_key:
        aws_access_key_id = settings.get("AWS_ACCESS_KEY_ID")
        aws_secret_access_key = settings.get("AWS_SECRET_ACCESS_KEY")

    return (aws_access_key_id, aws_secret_access_key)



import boto3

from spidermon.exceptions import NotConfigured
from spidermon.utils.settings import get_aws_credentials

from . import SendEmail


class SendSESEmail(SendEmail):
    aws_access_key = None
    aws_secret_key = None
    aws_region_name = "us-east-1"

    def __init__(
        self,
        aws_access_key=None,
        aws_secret_key=None,
        aws_region_name=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.aws_access_key = aws_access_key or self.aws_access_key
        self.aws_secret_key = aws_secret_key or self.aws_secret_key
        self.aws_region_name = aws_region_name or self.aws_region_name
        if not self.fake and not self.aws_access_key:
            raise NotConfigured("You must provide the AWS Access Key.")
        if not self.fake and not self.aws_secret_key:
            raise NotConfigured("You must provide the AWS Secret Key.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(
            crawler.settings
        )
        kwargs.update(
            {
                "aws_access_key": aws_access_key_id,
                "aws_secret_key": aws_secret_access_key,
                "aws_region_name": crawler.settings.get("SPIDERMON_AWS_REGION_NAME"),
            }
        )
        return kwargs

    def _get_recipients(self):
        recipients = []
        for recipient_subset in (self.to, self.cc, self.bcc):
            if not recipient_subset:
                pass
            elif isinstance(recipient_subset, str):
                recipients.append(recipient_subset)
            else:
                recipients.extend(recipient_subset)
        return recipients

    def send_message(self, message):
        client = boto3.client(
            service_name="ses",
            region_name=self.aws_region_name,
            aws_access_key_id=self.aws_access_key,
            aws_secret_access_key=self.aws_secret_key,
        )
        client.send_raw_email(
            Source=self.sender,
            Destinations=self._get_recipients(),
            RawMessage={"Data": message.as_string()},
        )



from spidermon.contrib.validation.translator import MessageTranslator
from spidermon.contrib.validation import messages


class JSONSchemaMessageTranslator(MessageTranslator):
    messages = {
        r"^.+ is a required property$": messages.MISSING_REQUIRED_FIELD,
        r"^.+ is not of type u?'array'$": messages.INVALID_ARRAY,
        r"^.+ is not of type u?'boolean'$": messages.INVALID_BOOLEAN,
        r"^.+ is not of type u?'integer'$": messages.INVALID_INT,
        r"^.+ is not of type u?'number'$": messages.INVALID_NUMBER,
        r"^.+ is not of type u?'object'$": messages.INVALID_OBJECT,
        r"^.+ is not of type u?'string'$": messages.INVALID_STRING,
        r"^.+ is not of type u?'null'$": messages.NOT_NULL,
        r"^.+ is not valid under any of the given schemas$": messages.NOT_VALID_UNDER_ANY_SCHEMA,
        r"^.+ is valid under each of .+$": messages.VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS,
        r"^Additional items are not allowed .*$": messages.TOO_MANY_ITEMS,
        r"^Additional properties are not allowed (?P<unexpected_fields>.*)$": messages.UNEXPECTED_FIELDS,
        r"^.+ is a dependency of .+$": messages.MISSING_DEPENDENT_FIELD,
        r"^.* is not one of .+$": messages.VALUE_NOT_IN_CHOICES,
        r"^.* is not a 'date-time'$": messages.INVALID_DATETIME,
        r"^.* is not a 'email'$": messages.INVALID_EMAIL,
        r"^.* is not a 'ipv4'$": messages.INVALID_IPV4,
        r"^.* is not a 'ipv6'$": messages.INVALID_IPV6,
        r"^.* is not a 'hostname'$": messages.INVALID_HOSTNAME,
        r"^.* is not a 'url'$": messages.INVALID_URL,
        r"^.* is not a 'uri'$": messages.INVALID_URI,
        r"^.* is not a 'regex'$": messages.INVALID_REGEX,
        r"^.* is not a 'color'$": messages.INVALID_COLOR,
        r"^.* has too many properties$": messages.TOO_MANY_PROPERTIES,
        r"^.* does not have enough properties": messages.NOT_ENOUGH_PROPERTIES,
        r"^.* is greater than the maximum of .*$": messages.NUMBER_TOO_HIGH,
        r"^.* is greater than or equal to the maximum of .*$": messages.NUMBER_TOO_HIGH,
        r"^.* is less than the minimum of .*$": messages.NUMBER_TOO_LOW,
        r"^.* is less than or equal to the minimum of .*$": messages.NUMBER_TOO_LOW,
        r"^.* is not a multiple of .*$": messages.NOT_MULTIPLE_OF,
        r"^.* is not allowed for .*$": messages.NOT_ALLOWED_VALUE,
        r"^.+ is too short$": messages.FIELD_TOO_SHORT,
        r"^.+ is too long$": messages.FIELD_TOO_LONG,
        r"^.+ does not match .*$": messages.REGEX_NOT_MATCHED,
        r"^.+ has non-unique elements$": messages.NOT_UNIQUE,
    }



from spidermon.core.actions import Action
from spidermon.exceptions import NotConfigured


class JobTagsAction(Action):
    tags = []
    tag_settings = None

    def __init__(self, tags=None):
        super().__init__()
        tags = tags or self.tags
        self.tags = tags if isinstance(tags, list) else [tags]

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        if cls.tag_settings:
            kwargs.update({"tags": crawler.settings.get(cls.tag_settings)})
        return kwargs

    def run_action(self):
        if self.tags:
            if not self.data.job:
                raise NotConfigured("Job not available!")
            job_metadata = self.data.job.metadata
            self.process_tags(job_metadata)
            job_metadata.save()

    def process_tags(self, job_metadata):
        raise NotImplementedError


class AddJobTags(JobTagsAction):
    tag_settings = "SPIDERMON_JOB_TAGS_TO_ADD"

    def process_tags(self, job_metadata):
        for tag in self.tags:
            if tag not in job_metadata["tags"]:
                job_metadata["tags"].append(tag)


class RemoveJobTags(JobTagsAction):
    tag_settings = "SPIDERMON_JOB_TAGS_TO_REMOVE"

    def process_tags(self, job_metadata):
        for tag in self.tags:
            if tag in job_metadata["tags"]:
                job_metadata["tags"].remove(tag)
